import logging
from dataclasses import dataclass
from enum import Enum, Flag, unique
from typing import List, Optional, Set, Tuple, Union

from asn1crypto import x509
from asn1crypto.x509 import KeyUsage
from .aurora_certvalidator.authority import AuthorityWithCert
from .aurora_certvalidator.errors import InvalidCertificateError
from .aurora_certvalidator.path import ValidationPath

from .aurora_signature.pdf_utils import generic
from .aurora_signature.pdf_utils.generic import pdf_name, pdf_string
from .aurora_signature.pdf_utils.misc import (
    OrderedEnum,
    PdfReadError,
    get_and_apply,
)
from .aurora_signature.sign.general import SigningError, UnacceptableSignerError
from .aurora_signature.pdf_utils.rw_common import PdfHandler
from .aurora_signature.pdf_utils.writer import BasePdfFileWriter
from .aurora_signature.pdf_utils.misc import (
    rd,
)

__all__ = [
    'SigFieldSpec',
    'SigSeedValueSpec',
    'MDPPerm',
    'SigAuthType',
    'SigSeedSubFilter',
    'enumerate_sig_fields'
]

logger = logging.getLogger(__name__)

class MDPPerm(OrderedEnum):
    """
    Indicates a ``/DocMDP`` level.

    Cf. Table 254  in ISO 32000-1.
    """

    NO_CHANGES = 1
    """
    No changes to the document are allowed.

    .. warning::
        This does not apply to DSS updates and the addition of document time
        stamps.
    """
    FILL_FORMS = 2
    """
    Form filling & signing is allowed.
    """

    ANNOTATE = 3
    """
    Form filling, signing and commenting are allowed.

    .. warning::
        Validating this ``/DocMDP`` level is not currently supported,
        but included in the list for completeness.
    """

@unique
class SigAuthType(Enum):
    """
    Enum declaring all supported ``/Prop_AuthType`` values.
    """

    PIN = pdf_string("PIN")
    PASSWORD = pdf_string("Password")
    FINGERPRINT = pdf_string("Fingerprint")


@unique
class SigSeedSubFilter(Enum):
    """
    Enum declaring all supported ``/SubFilter`` values.
    """

    ADOBE_PKCS7_DETACHED = pdf_name("/adbe.pkcs7.detached")
    PADES = pdf_name("/ETSI.CAdES.detached")
    ETSI_RFC3161 = pdf_name("/ETSI.RFC3161")

@dataclass(frozen=True)
class VisibleSigSettings:
    """
    .. versionadded:: 0.14.0

    Additional flags used when setting up visible signature widgets.
    """

    rotate_with_page: bool = True
    """
    Allow the signature widget to rotate with the page if rotation is applied
    (e.g. by way of the page's ``/Rotate`` entry). Default is ``True``.

    .. note::
        If ``False``, this will cause the ``NoRotate`` flag to be set.
    """

    scale_with_page_zoom: bool = True
    """
    Allow the signature widget to scale with the page's zoom level.
    Default is ``True``.

    .. note::
        If ``False``, this will cause the ``NoZoom`` flag to be set.
    """

    print_signature: bool = True
    """
    Render the signature when the document is printed. Default ``True``.
    """

class SigSeedValFlags(Flag):
    """
    Flags for the ``/Ff`` entry in the seed value dictionary for a signature
    field. These mark which of the constraints are to be strictly enforced,
    as opposed to optional ones.

    .. warning::
        The flags :attr:`LEGAL_ATTESTATION` and :attr:`APPEARANCE_FILTER` are
        processed in accordance with the specification when creating a
        signature, but support is nevertheless limited.

        * aurora does not support legal attestations at all, so given that
          the :attr:`LEGAL_ATTESTATION` requirement flag only restricts the
          legal attestations that can be used by the signer, aurora can safely
          ignore it when signing.

          On the other hand, since the validator is not aware of
          legal attestations either, it cannot validate signatures that
          make :attr:`~.SigSeedValueSpec.legal_attestations` a mandatory
          constraint.
        * Since aurora does not define any named appearances, setting
          the :attr:`APPEARANCE_FILTER` flag and the
          :attr:`~.SigSeedValueSpec.appearance` entry in the seed value
          dictionary will make aurora refuse to sign the document.

          When validating, the situation is different: since aurora has no
          way of knowing whether the signer used the named appearance imposed
          by the seed value dictionary, it will simply emit a warning and
          continue validating the signature.
    """

    FILTER = 1
    """
    Makes the signature handler setting mandatory. aurora only supports
    ``/Adobe.PPKLite``.
    """

    SUBFILTER = 2
    """
    See :attr:`~.SigSeedValueSpec.subfilters`.
    """

    V = 4
    """
    See :attr:`~.SigSeedValueSpec.sv_dict_version`.
    """

    REASONS = 8
    """
    See :attr:`~.SigSeedValueSpec.reasons`.
    """

    LEGAL_ATTESTATION = 16
    """
    See :attr:`~.SigSeedValueSpec.legal_attestations`.
    """

    ADD_REV_INFO = 32
    """
    See :attr:`~.SigSeedValueSpec.add_rev_info`.
    """

    DIGEST_METHOD = 64
    """
    See :attr:`~.SigSeedValueSpec.digest_method`.
    """

    LOCK_DOCUMENT = 128
    """
    See :attr:`~.SigSeedValueSpec.lock_document`.
    """

    APPEARANCE_FILTER = 256
    """
    See :attr:`~.SigSeedValueSpec.appearance`.
    """

class SigCertConstraintFlags(Flag):
    """
    Flags for the ``/Ff`` entry in the certificate seed value dictionary for
    a dictionary field. These mark which of the constraints are to be
    strictly enforced, as opposed to optional ones.

    .. warning::
        While this enum records values for all flags, not all corresponding
        constraint types have been implemented yet.
    """

    SUBJECT = 1
    """
    See :attr:`SigCertConstraints.subjects`.
    """

    ISSUER = 2
    """
    See :attr:`SigCertConstraints.issuers`.
    """

    OID = 4
    """
    Currently not supported.
    """

    SUBJECT_DN = 8
    """ 
    See :attr:`SigCertConstraints.subject_dn`.
    """

    RESERVED = 16
    """
    Currently not supported (reserved).
    """

    KEY_USAGE = 32
    """
    See :attr:`SigCertConstraints.key_usage`.
    """

    URL = 64
    """
    See :attr:`SigCertConstraints.info_url`.
    
    .. note::
        As specified in the standard, this enforcement bit is supposed to be
        ignored by default. We include it for compatibility reasons.
    """

    UNSUPPORTED = RESERVED | OID
    """
    Flags for which the corresponding constraint is unsupported.
    """


class SigCertKeyUsage:
    """
    Encodes the key usage bits that must (resp. must not) be active on the
    signer's certificate.

    .. note::
        See § 4.2.1.3 in :rfc:`5280` and :class:`.KeyUsage` for more
        information on key usage extensions.

    .. note::
        The human-readable names of the key usage extensions are recorded
        in ``camelCase`` in :rfc:`5280`, but this class uses
        the naming convention of :class:`.KeyUsage` in ``asn1crypto``.
        The conversion is done by replacing ``camelCase`` with ``snake_case``.
        For example, ``nonRepudiation`` becomes ``non_repudiation``, and
        ``digitalSignature`` turns into ``digital_signature``.

    .. note::
        This class is intended to closely replicate the definition of the
        KeyUsage entry Table 235 in ISO 32000-1.
        In particular, it does *not* provide a mechanism to deal
        with extended key usage extensions (cf. § 4.2.1.12 in :rfc:`5280`).

    :param must_have:
        The :class:`.KeyUsage` object encoding the key usage extensions
        that must be present on the signer's certificate.
    :param forbidden:
        The :class:`.KeyUsage` object encoding the key usage extensions
        that must *not* be present on the signer's certificate.
    """

    def __init__(
        self,
        must_have: Optional[KeyUsage] = None,
        forbidden: Optional[KeyUsage] = None,
    ):
        self.must_have = must_have if must_have is not None else KeyUsage(set())
        self.forbidden = forbidden if forbidden is not None else KeyUsage(set())

    def encode_to_sv_string(self):
        """
        Encode the key usage requirements in the format specified in the PDF
        specification.

        :return:
            A string.
        """

        def fmt_bit(bit: int):
            if self.must_have[bit]:
                return '1'
            elif self.forbidden[bit]:
                return '0'
            else:
                return 'X'

        return ''.join(fmt_bit(bit) for bit in range(9))

    @classmethod
    def read_from_sv_string(cls, ku_str):
        """
        Parse a PDF KeyUsage string into an instance of
        :class:`.SigCertKeyUsage`. See Table 235 in ISO 32000-1.

        :param ku_str:
            A PDF KeyUsage string.
        :return:
            An instance of :class:`.SigCertKeyUsage`.
        """
        ku_str = ku_str[:9]

        def _as_tuple(with_val):
            return tuple(1 if val == with_val else 0 for val in ku_str)

        return SigCertKeyUsage(
            must_have=KeyUsage(_as_tuple('1')),
            forbidden=KeyUsage(_as_tuple('0')),
        )

    @classmethod
    def from_sets(
        cls,
        must_have: Optional[Set[str]] = None,
        forbidden: Optional[Set[str]] = None,
    ):
        """
        Initialise a :class:`.SigCertKeyUsage` object from two sets.

        :param must_have:
            The key usage extensions that must be present on the signer's
            certificate.
        :param forbidden:
            The key usage extensions that must *not* be present on the signer's
            certificate.
        :return:
            A :class:`.SigCertKeyUsage` object encoding these.
        """
        return SigCertKeyUsage(
            must_have=KeyUsage(set() if must_have is None else must_have),
            forbidden=KeyUsage(set() if forbidden is None else forbidden),
        )

    def must_have_set(self) -> Set[str]:
        """
        Return the set of key usage extensions that must be present
        on the signer's certificate.
        """
        return self.must_have.native

    def forbidden_set(self) -> Set[str]:
        """
        Return the set of key usage extensions that must not be present
        on the signer's certificate.
        """
        return self.forbidden.native

    def __eq__(self, other):
        return (
            isinstance(other, SigCertKeyUsage)
            and self.must_have_set() == other.must_have_set()
            and self.forbidden_set() == other.forbidden_set()
        )

name_type_abbrevs = {
    '2.5.4.3': 'CN',
    '2.5.4.5': 'SerialNumber',
    '2.5.4.6': 'C',
    '2.5.4.7': 'L',
    '2.5.4.8': 'ST',
    '2.5.4.10': 'O',
    '2.5.4.11': 'OU',
}
name_type_abbrevs_rev = {v: k for k, v in name_type_abbrevs.items()}

def x509_name_keyval_pairs(name: x509.Name, abbreviate_oids=False):
    rdns: x509.RDNSequence = name.chosen
    for rdn in rdns:
        for type_and_value in rdn:
            oid: x509.NameType = type_and_value['type']
            # these are all some kind of string, and the PDF
            # standard says that the value should be a text string object,
            # so we just have asn1crypto convert everything to strings
            value = type_and_value['value']
            key = oid.dotted
            if abbreviate_oids:
                key = name_type_abbrevs.get(key, key)

            yield key, value.native
            # these should be strings

@dataclass(frozen=True)
class SigCertConstraints:
    """
    This part of the seed value dictionary allows the document author
    to set constraints on the signer's certificate.

    See Table 235 in ISO 32000-1.
    """

    flags: SigCertConstraintFlags = SigCertConstraintFlags(0)
    """
    Enforcement flags. By default, all entries are optional.
    """

    subjects: Optional[List[x509.Certificate]] = None
    """
    Explicit list of certificates that can be used to sign a signature field.
    """

    subject_dn: Optional[x509.Name] = None
    """
    Certificate subject names that can be used to sign a signature field.
    Subject DN entries that are not mentioned are unconstrained.
    """

    issuers: Optional[List[x509.Certificate]] = None
    """
    List of issuer certificates that the signer certificate can be issued by.
    Note that these issuers do not need to be the *direct* issuer of the
    signer's certificate; any descendant relationship will do.
    """

    info_url: Optional[str] = None
    """
    Informational URL that should be opened when an appropriate certificate
    cannot be found (if :attr:`url_type` is ``/Browser``, that is).
    
    .. note::
        aurora ignores this value, but we include it for compatibility.
    """

    url_type: generic.NameObject = pdf_name('/Browser')
    """
    Handler that should be used to open :attr:`info_url`.
    ``/Browser`` is the only implementation-independent value.
    """

    key_usage: Optional[List[SigCertKeyUsage]] = None
    """
    Specify the key usage extensions that should (or should not) be present
    on the signer's certificate.
    """

    # TODO support OID constraints (certificate policies) and signature policy
    #  constraints.

    @classmethod
    def from_pdf_object(cls, pdf_dict):
        """
        Read a PDF dictionary into a :class:`.SigCertConstraints` object.

        :param pdf_dict:
            A :class:`~.generic.DictionaryObject`.
        :return:
            A :class:`.SigCertConstraints` object.
        """

        if isinstance(pdf_dict, generic.IndirectObject):
            pdf_dict = pdf_dict.get_object()
        try:
            if pdf_dict['/Type'] != '/SVCert':  # pragma: nocover
                raise PdfReadError('Object /Type entry is not /SVCert')
        except KeyError:  # pragma: nocover
            pass
        flags = SigCertConstraintFlags(pdf_dict.get('/Ff', 0))
        subjects = [
            x509.Certificate.load(cert.original_bytes)
            for cert in pdf_dict.get('/Subject', ())
        ]
        issuers = [
            x509.Certificate.load(cert.original_bytes)
            for cert in pdf_dict.get('/Issuer', ())
        ]

        def format_attr(attr):
            # strip initial /
            attr = attr[1:]
            # attempt to convert abbreviated attrs to OIDs, since build()
            # takes OIDs
            return name_type_abbrevs_rev.get(attr.upper(), attr)

        subject_dns = x509.Name.build(
            {
                format_attr(attr): value
                for dn_dir in pdf_dict.get('/SubjectDN', ())
                for attr, value in dn_dir.items()
            }
        )

        def parse_key_usage(val):
            return [SigCertKeyUsage.read_from_sv_string(ku) for ku in val]

        key_usage = get_and_apply(pdf_dict, '/KeyUsage', parse_key_usage)

        url = pdf_dict.get('/URL')
        url_type = pdf_dict.get('/URLType')
        kwargs = {
            'flags': flags,
            'subjects': subjects or None,
            'subject_dn': subject_dns or None,
            'issuers': issuers or None,
            'info_url': url,
            'key_usage': key_usage,
        }
        if url is not None and url_type is not None:
            kwargs['url_type'] = url_type
        return cls(**kwargs)

    def as_pdf_object(self):
        """
        Render this :class:`.SigCertConstraints` object to a PDF dictionary.

        :return:
            A :class:`~.generic.DictionaryObject`.
        """

        result = generic.DictionaryObject(
            {
                pdf_name('/Type'): pdf_name('/SVCert'),
                pdf_name('/Ff'): generic.NumberObject(self.flags.value),
            }
        )
        if self.subjects is not None:
            result[pdf_name('/Subject')] = generic.ArrayObject(
                generic.ByteStringObject(cert.dump()) for cert in self.subjects
            )
        if self.subject_dn:
            # FIXME Adobe Reader seems to ignore this for some reason.
            #  Should try to figure out what I'm doing wrong
            result[pdf_name('/SubjectDN')] = generic.ArrayObject(
                [
                    generic.DictionaryObject(
                        {
                            pdf_name('/' + key): pdf_string(value)
                            for key, value in x509_name_keyval_pairs(
                                self.subject_dn, abbreviate_oids=True
                            )
                        }
                    )
                ]
            )
        if self.issuers is not None:
            result[pdf_name('/Issuer')] = generic.ArrayObject(
                generic.ByteStringObject(cert.dump()) for cert in self.issuers
            )
        if self.info_url is not None:
            result[pdf_name('/URL')] = pdf_string(self.info_url)
            result[pdf_name('/URLType')] = self.url_type

        if self.key_usage is not None:
            result[pdf_name('/KeyUsage')] = generic.ArrayObject(
                pdf_string(ku.encode_to_sv_string()) for ku in self.key_usage
            )

        return result

    def satisfied_by(
        self,
        signer: x509.Certificate,
        validation_path: Optional[ValidationPath],
    ):
        """
        Evaluate whether a signing certificate satisfies the required
        constraints of this :class:`.SigCertConstraints` object.

        :param signer:
            The candidate signer's certificate.
        :param validation_path:
            Validation path of the signer's certificate.
        :raises UnacceptableSignerError:
            Raised if the conditions are not met.
        """
        # this function assumes that key usage & trust checks have
        #  passed already.
        flags = self.flags
        if flags & SigCertConstraintFlags.UNSUPPORTED:
            raise NotImplementedError(
                "Certificate constraint flags include mandatory constraints "
                "that are not supported."
            )
        if (
            flags & SigCertConstraintFlags.SUBJECT
        ) and self.subjects is not None:
            # Explicit whitelist of approved signer certificates
            # compare using issuer_serial
            acceptable = (s.issuer_serial for s in self.subjects)
            if signer.issuer_serial not in acceptable:
                raise UnacceptableSignerError(
                    "Signer certificate not on SVCert whitelist."
                )
        if (flags & SigCertConstraintFlags.ISSUER) and self.issuers is not None:
            if validation_path is None:
                raise UnacceptableSignerError("Validation path not provided.")
            # Here, we need to match any issuer in the chain of trust to
            #  any of the issuers on the approved list.

            # To do so, we collect all issuer_serial identifiers in the chain
            # for all certificates except the last one (i.e. the current signer)
            path_iss_serials = {
                authority.certificate.issuer_serial
                for authority in validation_path.iter_authorities()
                if isinstance(authority, AuthorityWithCert)
            }
            for issuer in self.issuers:
                if issuer.issuer_serial in path_iss_serials:
                    break
            else:
                # raise error if the loop runs to completion
                raise UnacceptableSignerError(
                    "Signer certificate cannot be traced back to approved "
                    "issuer."
                )
        if (flags & SigCertConstraintFlags.SUBJECT_DN) and self.subject_dn:
            # I'm not entirely sure whether my reading of the standard is
            #  is correct, but I believe that this is the intention:
            # A DistinguishedName object is a sequence of
            #  relative distinguished names (RDNs). The contents of the
            #  /SubjectDN specify a list of constraints that might apply to each
            #  of these RDNs. I believe the requirement is that each of the
            #  SubjectDN entries must match one of these RDNs.

            requirement_list = list(x509_name_keyval_pairs(self.subject_dn))
            subject_name = list(x509_name_keyval_pairs(signer.subject))
            if not all(attr in subject_name for attr in requirement_list):
                raise UnacceptableSignerError(
                    "Subject does not have some of the following required "
                    "attributes: " + self.subject_dn.human_friendly
                )
        if (
            flags & SigCertConstraintFlags.KEY_USAGE
        ) and self.key_usage is not None:
            from .validation.settings import KeyUsageConstraints

            for ku in self.key_usage:
                try:
                    KeyUsageConstraints(
                        key_usage=ku.must_have_set(),
                        key_usage_forbidden=ku.forbidden_set(),
                        # This is the way ISO 32k does things
                        match_all_key_usages=True,
                    ).validate(signer)
                    break
                except InvalidCertificateError:
                    continue
            else:
                raise UnacceptableSignerError(
                    "The signer satisfies none of the key usage "
                    "extension profiles specified in the seed value dictionary."
                )

class SeedSignatureType:
    """
    Signature type indicator to be embedded into the seed value dictionary
    attached to a signature field.

    :param mdp_perm:
        If not ``None``, indicates that the signature field is intended for
        a certification signature. The :class:`MDPPerm` value passed as the
        ``mdp_perm`` parameter indicates the modification policy that the
        certification signature should use.

        A value of ``None`` indicates that the signature field is intended for
        an approval signature (i.e. a non-certification signature).
    """

    def __init__(self, mdp_perm: Optional[MDPPerm] = None):
        self.mdp_perm = mdp_perm

    def __eq__(self, other):
        return (
            isinstance(other, SeedSignatureType)
            and other.mdp_perm == self.mdp_perm
        )

    def certification_signature(self) -> bool:
        return self.mdp_perm is not None

@unique
class SeedValueDictVersion(OrderedEnum):
    """
    Specify the minimal compliance level for a seed value dictionary processor.
    """

    PDF_1_5 = 1
    """
    Require the reader to understand all keys defined in PDF 1.5.
    """

    PDF_1_7 = 2
    """
    Require the reader to understand all keys defined in PDF 1.7.
    """

    PDF_2_0 = 3
    """
    Require the reader to understand all keys defined in PDF 2.0.
    """

@unique
class SeedLockDocument(Enum):
    """
    Provides a recommendation to the signer as to whether the document should
    be locked after signing.
    The corresponding flag in :attr:`.SigSeedValueSpec.flags` determines whether
    this constraint is a required constraint.
    """

    LOCK = pdf_name('/true')
    """
    Lock the document after signing.
    """

    DO_NOT_LOCK = pdf_name('/false')
    """
    Lock the document after signing.
    """

    SIGNER_DISCRETION = pdf_name('/auto')
    """
    Leave the decision up to the signer.
    
    .. note::
        This is functionally equivalent to not specifying any value.
    """


@dataclass(frozen=True)
class SigSeedValueSpec:
    """
    Python representation of a PDF seed value dictionary.
    """

    flags: SigSeedValFlags = SigSeedValFlags(0)
    """
    Enforcement flags. By default, all entries are optional.
    """

    reasons: Optional[List[str]] = None
    """
    Acceptable reasons for signing.
    """

    timestamp_server_url: Optional[str] = None
    """
    RFC 3161 timestamp server endpoint suggestion.
    """

    timestamp_required: bool = False
    """
    Flags whether a timestamp is required.
    This flag is only meaningful if :attr:`timestamp_server_url` is specified.
    """

    cert: Optional[SigCertConstraints] = None
    """
    Constraints on the signer's certificate.
    """

    subfilters: Optional[List[SigSeedSubFilter]] = None
    """
    Acceptable ``/SubFilter`` values.
    """

    digest_methods: Optional[List[str]] = None
    """
    Acceptable digest methods.
    """

    add_rev_info: Optional[bool] = None
    """
    Indicates whether revocation information should be embedded.
    
    .. warning::
        This flag exclusively refers to the Adobe-style revocation information
        embedded within the CMS object that is written to the signature field.
        PAdES-style revocation information that is saved to the document
        security store (DSS) does *not* satisfy the requirement.
        Additionally, the standard mandates that ``/SubFilter`` be equal to
        ``/adbe.pkcs7.detached`` if this flag is ``True``.
    """

    seed_signature_type: Optional[SeedSignatureType] = None
    """
    Specifies the type of signature that should occupy a signature field;
    this represents the ``/MDP`` entry in the seed value dictionary.
    See :class:`.SeedSignatureType` for details.
    
    .. caution::
        Since a certification-type signature is by definition the first
        signature applied to a document, compliance with this requirement
        cannot be cryptographically enforced.
    """

    sv_dict_version: Union[SeedValueDictVersion, int, None] = None
    """
    Specifies the compliance level required of a seed value dictionary
    processor. If ``None``, aurora will compute an appropriate value.
    
    .. note::
        You may also specify this value directly as an integer.
        This covers potential future versions of the standard that aurora
        does not support out of the box.
    """

    legal_attestations: Optional[List[str]] = None
    """
    Specifies the possible legal attestations that a certification signature
    occupying this signature field can supply.
    The corresponding flag in :attr:`flags` indicates whether this is a
    mandatory constraint.
    
    .. caution::
        Since :attr:`legal_attestations` is only relevant for certification
        signatures, compliance with this requirement cannot be reliably 
        enforced.
        Regardless, since aurora's validator is also unaware of legal
        attestation settings, it will refuse to validate signatures
        where this seed value constitutes a mandatory constraint.
        
        Additionally, since aurora does not support legal attestation
        specifications at all, it vacuously satisfies the requirements of this
        entry no matter what, and will therefore ignore it when signing.
    """

    lock_document: Optional[SeedLockDocument] = None
    """
    Tell the signer whether or not the document should be locked after signing
    this field; see :class:`.SeedLockDocument` for details.
    
    The corresponding flag in :attr:`flags` indicates whether this constraint
    is mandatory.
    """

    # TODO handle this value by reading named appearances from the user's
    #  settings

    appearance: Optional[str] = None
    """
    Specify a named appearance to use when generating the signature.
    The corresponding flag in :attr:`flags` indicates whether this constraint
    is mandatory.
    
    .. caution::
        There is no standard registry of named appearances, so these constraints
        are not portable, and cannot be validated.
        
        aurora currently does not define any named appearances.        
    """

    def as_pdf_object(self):
        """
        Render this :class:`.SigSeedValueSpec` object to a PDF dictionary.

        :return:
            A :class:`~.generic.DictionaryObject`.
        """
        min_version = SeedValueDictVersion.PDF_1_5
        result = generic.DictionaryObject(
            {
                pdf_name('/Type'): pdf_name('/SV'),
                pdf_name('/Ff'): generic.NumberObject(self.flags.value),
            }
        )

        if self.subfilters is not None:
            result[pdf_name('/SubFilter')] = generic.ArrayObject(
                sf.value for sf in self.subfilters
            )
        if self.add_rev_info is not None:
            min_version = SeedValueDictVersion.PDF_1_7
            result[pdf_name('/AddRevInfo')] = generic.BooleanObject(
                self.add_rev_info
            )
        if self.digest_methods is not None:
            min_version = SeedValueDictVersion.PDF_1_7
            result[pdf_name('/DigestMethod')] = generic.ArrayObject(
                map(pdf_string, self.digest_methods)
            )
        if self.reasons is not None:
            result[pdf_name('/Reasons')] = generic.ArrayObject(
                pdf_string(reason) for reason in self.reasons
            )
        if self.timestamp_server_url is not None:
            min_version = SeedValueDictVersion.PDF_1_7
            result[pdf_name('/TimeStamp')] = generic.DictionaryObject(
                {
                    pdf_name('/URL'): pdf_string(self.timestamp_server_url),
                    pdf_name('/Ff'): generic.NumberObject(
                        1 if self.timestamp_required else 0
                    ),
                }
            )
        if self.cert is not None:
            result[pdf_name('/Cert')] = self.cert.as_pdf_object()
        if self.seed_signature_type is not None:
            mdp_perm = self.seed_signature_type.mdp_perm
            result[pdf_name('/MDP')] = generic.DictionaryObject(
                {
                    pdf_name('/P'): generic.NumberObject(
                        mdp_perm.value if mdp_perm is not None else 0
                    )
                }
            )
        if self.legal_attestations is not None:
            result[pdf_name('/LegalAttestation')] = generic.ArrayObject(
                pdf_string(att) for att in self.legal_attestations
            )
        if self.lock_document is not None:
            min_version = SeedValueDictVersion.PDF_2_0
            result[pdf_name('/LockDocument')] = self.lock_document.value
        if self.appearance is not None:
            result[pdf_name('/AppearanceFilter')] = pdf_string(self.appearance)

        specified_version = self.sv_dict_version
        if specified_version is not None:
            result[pdf_name('/V')] = generic.NumberObject(
                specified_version.value
                if isinstance(specified_version, SeedValueDictVersion)
                else specified_version
            )
        else:
            result[pdf_name('/V')] = generic.NumberObject(min_version.value)
        return result

    @classmethod
    def from_pdf_object(cls, pdf_dict):
        """
        Read from a seed value dictionary.

        :param pdf_dict:
            A :class:`~.generic.DictionaryObject`.
        :return:
            A :class:`.SigSeedValueSpec` object.
        """
        if isinstance(pdf_dict, generic.IndirectObject):
            pdf_dict = pdf_dict.get_object()
        try:
            if pdf_dict['/Type'] != '/SV':  # pragma: nocover
                raise PdfReadError('Object /Type entry is not /SV')
        except KeyError:  # pragma: nocover
            pass

        flags = SigSeedValFlags(pdf_dict.get('/Ff', 0))
        try:
            sig_filter = pdf_dict['/Filter']
            if (flags & SigSeedValFlags.FILTER) and (
                sig_filter != '/Adobe.PPKLite'
            ):
                raise SigningError(
                    "Signature handler '%s' is not available, only the "
                    "default /Adobe.PPKLite is supported." % sig_filter
                )
        except KeyError:
            pass

        try:
            min_version = pdf_dict['/V']
            supported = SeedValueDictVersion.PDF_2_0.value
            if flags & SigSeedValFlags.V and min_version > supported:
                raise SigningError(
                    "Seed value dictionary version %s not supported."
                    % min_version
                )
            min_version = SeedValueDictVersion(min_version)
        except KeyError:
            min_version = None

        try:
            add_rev_info = bool(pdf_dict['/AddRevInfo'])
        except KeyError:
            add_rev_info = None

        subfilter_reqs = pdf_dict.get('/SubFilter', None)
        subfilters = None
        if subfilter_reqs is not None:

            def _subfilters():
                for s in subfilter_reqs:
                    try:
                        yield SigSeedSubFilter(s)
                    except ValueError:
                        pass

            subfilters = list(_subfilters())

        try:
            digest_methods = [s.lower() for s in pdf_dict['/DigestMethod']]
        except KeyError:
            digest_methods = None

        reasons = get_and_apply(pdf_dict, '/Reasons', list)
        legal_attestations = get_and_apply(pdf_dict, '/LegalAttestation', list)

        def read_mdp_dict(mdp):
            try:
                val = mdp['/P']
                return SeedSignatureType(None if val == 0 else MDPPerm(val))
            except (KeyError, TypeError, ValueError):
                raise SigningError(
                    f"/MDP entry {mdp} in seed value dictionary is not "
                    "correctly formatted."
                )

        signature_type = get_and_apply(pdf_dict, '/MDP', read_mdp_dict)

        def read_lock_document(val):
            try:
                return SeedLockDocument(val)
            except ValueError:
                raise SigningError(f"/LockDocument entry '{val}' is invalid.")

        lock_document = get_and_apply(
            pdf_dict, '/LockDocument', read_lock_document
        )
        appearance_filter = pdf_dict.get('/AppearanceFilter', None)
        timestamp_dict = pdf_dict.get('/TimeStamp', {})
        timestamp_server_url = timestamp_dict.get('/URL', None)
        timestamp_required = bool(timestamp_dict.get('/Ff', 0))
        cert_constraints = pdf_dict.get('/Cert', None)
        if cert_constraints is not None:
            cert_constraints = SigCertConstraints.from_pdf_object(
                cert_constraints
            )
        return cls(
            flags=flags,
            reasons=reasons,
            timestamp_server_url=timestamp_server_url,
            cert=cert_constraints,
            subfilters=subfilters,
            digest_methods=digest_methods,
            add_rev_info=add_rev_info,
            timestamp_required=timestamp_required,
            legal_attestations=legal_attestations,
            seed_signature_type=signature_type,
            sv_dict_version=min_version,
            lock_document=lock_document,
            appearance=appearance_filter,
        )

    def build_timestamper(self):
        """
        Return a timestamper object based on the :attr:`timestamp_server_url`
        attribute of this :class:`.SigSeedValueSpec` object.

        :return:
            A :class:`~.aurora_signature.sign.timestamps.HTTPTimeStamper`.
        """
        from .aurora_signature.sign.timestamps import HTTPTimeStamper

        if self.timestamp_server_url:
            return HTTPTimeStamper(self.timestamp_server_url)

class FieldMDPAction(Enum):
    """
    Marker for the scope of a ``/FieldMDP`` policy.
    """

    ALL = pdf_name('/All')
    """
    The policy locks all form fields.
    """

    INCLUDE = pdf_name('/Include')
    """
    The policy locks all fields in the list (see :attr:`.FieldMDPSpec.fields`).
    """

    EXCLUDE = pdf_name('/Exclude')
    """
    The policy locks all fields except those specified in the list
    (see :attr:`.FieldMDPSpec.fields`).
    """


@dataclass(frozen=True)
class FieldMDPSpec:
    """``/FieldMDP`` policy description.

    This class models both field lock dictionaries and ``/FieldMDP``
    transformation parameters.
    """

    action: FieldMDPAction
    """
    Indicates the scope of the policy.
    """

    fields: Optional[List[str]] = None
    """
    Indicates the fields subject to the policy,
    unless :attr:`action` is :attr:`.FieldMDPAction.ALL`.
    """

    def as_pdf_object(self) -> generic.DictionaryObject:
        """
        Render this ``/FieldMDP`` policy description as a PDF dictionary.

        :return:
            A :class:`~.generic.DictionaryObject`.
        """
        result = generic.DictionaryObject(
            {
                pdf_name('/Action'): self.action.value,
            }
        )
        if self.action != FieldMDPAction.ALL:
            result['/Fields'] = generic.ArrayObject(
                map(pdf_string, self.fields or ())
            )
        return result

    def as_transform_params(self) -> generic.DictionaryObject:
        """
        Render this ``/FieldMDP`` policy description as a PDF dictionary,
        ready for inclusion into the ``/TransformParams`` entry of a
        ``/FieldMDP`` dictionary associated with a signature object.

        :return:
            A :class:`~.generic.DictionaryObject`.
        """

        result = self.as_pdf_object()
        result['/Type'] = pdf_name('/TransformParams')
        result['/V'] = pdf_name('/1.2')
        return result

    def as_sig_field_lock(self) -> generic.DictionaryObject:
        """
        Render this ``/FieldMDP`` policy description as a PDF dictionary,
        ready for inclusion into the ``/Lock`` dictionary of a signature field.

        :return:
            A :class:`~.generic.DictionaryObject`.
        """

        result = self.as_pdf_object()
        result['/Type'] = pdf_name('/SigFieldLock')
        return result

    @classmethod
    def from_pdf_object(cls, pdf_dict) -> 'FieldMDPSpec':
        """
        Read a PDF dictionary into a :class:`.FieldMDPSpec` object.

        :param pdf_dict:
            A :class:`~.generic.DictionaryObject`.
        :return:
            A :class:`.FieldMDPSpec` object.
        """
        try:
            action = FieldMDPAction(pdf_dict['/Action'])
        except KeyError:  # pragma: nocover
            raise PdfReadError("/Action is required.")

        if action != FieldMDPAction.ALL:
            try:
                fields = pdf_dict['/Fields']
            except KeyError:  # pragma: nocover
                raise PdfReadError(
                    "/Fields is required when /Action is not /All"
                )
        else:
            fields = None
        return cls(action=action, fields=fields)

    def is_locked(self, field_name: str) -> bool:
        """
        Adjudicate whether a field should be locked by the policy described by
        this :class:`.FieldMDPSpec` object.

        :param field_name:
            The name of a form field.
        :return:
            ``True`` if the field should be locked, ``False`` otherwise.
        """
        if self.action == FieldMDPAction.ALL:
            return True

        lock_result = self.action == FieldMDPAction.INCLUDE
        for scoped_field_name in self.fields or ():
            # treat non-terminal field in/exclusions as including the whole
            # tree beneath them
            if field_name.startswith(scoped_field_name):
                return lock_result
        return not lock_result

@dataclass(frozen=True)
class InvisSigSettings:
    """
    Invisible signature widget generation settings.

    These settings exist because there is no real way of including an untagged
    invisible signature in a document that complies with the requirements
    of both PDF/A-2 (or -3) and PDF/UA-1.

    Compatibility with PDF/A (the default) requires the print flag to be set.
    Compatibility with PDF/UA requires the hidden flag to be set (which is
    banned in PDF/A) or the box to be outside the crop box.
    """

    set_print_flag: bool = True
    """
    Set the print flag. Required in PDF/A.
    """

    set_hidden_flag: bool = False
    """
    Set the hidden flag. Required in PDF/UA.
    """

    box_out_of_bounds: bool = False
    """
    Put the box out of bounds (technically, this just makes the box
    zero-sized with large negative coordinates).

    This is a hack to get around the fact that PDF/UA requires the hidden
    flag to be set on all in-bounds untagged annotations, and some validators
    consider [0, 0, 0, 0] to be an in-bounds rectangle if (0, 0) is a point
    that falls within the crop box.
    """


@dataclass(frozen=True)
class SigFieldSpec:
    """Description of a signature field to be created."""

    sig_field_name: str
    """
    Name of the signature field.
    """

    on_page: int = 0
    """
    Index of the page on which the signature field should be included (starting
    at `0`).
    A negative number counts pages from the back of the document,
    with index ``-1`` referring to the last page.
    
    .. note::
        This is essentially only relevant for visible signature fields, i.e.
        those that have a widget associated with them.
    """

    box: Optional[Tuple[int, int, int, int]] = None
    """
    Bounding box of the signature field, if applicable.

    Typically specified in ``ll_x``, ``ll_y``, ``ur_x``, ``ur_y`` format,
    where ``ll_*`` refers to the lower left and ``ur_*`` to the upper right
    corner.
    """

    seed_value_dict: Optional[SigSeedValueSpec] = None
    """
    Specification for the seed value dictionary, if applicable.
    """

    field_mdp_spec: Optional[FieldMDPSpec] = None
    """
    Specification for the field lock dictionary, if applicable.
    """

    doc_mdp_update_value: Optional[MDPPerm] = None
    """
    Value to use for the document modification policy associated with the
    signature in this field.
    
    This value will be embedded into the field lock dictionary if specified, and    
    is meaningless if :attr:`field_mdp_spec` is not specified.
    
    .. warning::
        DocMDP entries for approval signatures are a PDF 2.0 feature.
        Older PDF software will likely ignore this part of the field lock
        dictionary.
    """
    # TODO add a reference to the docs on certification once those are written.

    combine_annotation: bool = True
    """
    Flag controlling whether the field should be combined with its
    annotation dictionary; ``True`` by default.
    """

    empty_field_appearance: bool = False
    """
    Generate a neutral appearance stream for empty, visible signature fields.
    If ``False``, an empty appearance stream will be put in.
    
    .. note::
        We use an empty appearance stream to satisfy the appearance requirements
        for widget annotations in ISO 32000-2. However, even when a nontrivial
        appearance stream is present on an empty signature field, many viewers 
        will not use it to render the appearance of the empty field on-screen.

        Instead, these viewers typically substitute their own native widget.
    """

    invis_sig_settings: InvisSigSettings = InvisSigSettings()
    """
    Advanced settings to control invisible signature field generation.
    """

    readable_field_name: Optional[str] = None
    """
    Human-readable field name (``/TU`` entry).
    
    .. note::
        This value is commonly rendered as a tooltip in viewers, but also
        serves an accessibility purpose.
    """

    visible_sig_settings: VisibleSigSettings = VisibleSigSettings()
    """
    Advanced settings to control the generation of visible signature fields.
    """

    def format_lock_dictionary(self) -> Optional[generic.DictionaryObject]:
        if self.field_mdp_spec is None:
            return None
        result = self.field_mdp_spec.as_sig_field_lock()
        # this requires PDF 2.0 in principle, but meh, noncompliant
        # readers will ignore it anyway
        if self.doc_mdp_update_value is not None:
            result['/P'] = generic.NumberObject(self.doc_mdp_update_value.value)
        return result

def _insert_or_get_field_at(
    writer: BasePdfFileWriter,
    fields,
    path,
    parent_ref=None,
    modified=False,
    field_obj=None,
):
    current_partial, tail = path[0], path[1:]

    for field_ref in fields:
        assert isinstance(field_ref, generic.IndirectObject)
        field = field_ref.get_object()
        if field.get('/T', None) == current_partial:
            break
    else:
        # have to insert a new element into the fields array
        if field_obj is not None and not tail:
            field = field_obj
        else:
            # create a generic field
            field = generic.DictionaryObject()
        field['/T'] = pdf_string(current_partial)
        if parent_ref is not None:
            field['/Parent'] = parent_ref
        field_ref = writer.add_object(field)
        fields.append(field_ref)
        writer.update_container(fields)
        modified = True

    if not tail:
        return modified, field_ref
    # check for /Kids, and create it if necessary
    try:
        kids = field['/Kids']
    except KeyError:
        kids = field['/Kids'] = generic.ArrayObject()
        writer.update_container(field)
        modified = True

    # recurse in to /Kids array
    return _insert_or_get_field_at(
        writer,
        kids,
        tail,
        parent_ref=field_ref,
        modified=modified,
        field_obj=field_obj,
    )

def annot_width_height(
    annot_dict: generic.DictionaryObject,
) -> Tuple[float, float]:
    """
    Internal function to compute the width and height of an annotation.

    :param annot_dict:
        Annotation dictionary.
    :return:
        a (width, height) tuple
    """
    try:
        x1, y1, x2, y2 = annot_dict['/Rect']
    except KeyError:
        return 0, 0
    w = abs(x1 - x2)
    h = abs(y1 - y2)
    return w, h

def enumerate_sig_fields(
    handler: PdfHandler,
    filled_status: Optional[bool] = None,
    with_name: Optional[str] = None,
):
    """
    Enumerate signature fields.

    :param handler:
        The :class:`~.rw_common.PdfHandler` to operate on.
    :param filled_status:
        Optional boolean. If ``True`` (resp. ``False``) then all filled
        (resp. empty) fields are returned. If left ``None`` (the default), then
        all fields are returned.
    :param with_name:
        If not ``None``, only look for fields with the specified name.
    :return:
        A generator producing signature fields.
    """

    try:
        fields = handler.root['/AcroForm']['/Fields']
    except KeyError:
        return

    yield from enumerate_sig_fields_in(
        fields,
        filled_status=filled_status,
        with_name=with_name,
        refs_seen=set(),
    )

def enumerate_sig_fields_in(
    field_list,
    filled_status=None,
    with_name=None,
    parent_name="",
    parents=None,
    *,
    refs_seen,
):
    if not isinstance(field_list, generic.ArrayObject):
        logger.warning(
            f"Values of type {type(field_list)} are not valid as field "
            f"lists, must be array objects -- skipping."
        )
        return

    parents = parents or ()
    for field_ref in field_list:
        if not isinstance(field_ref, generic.IndirectObject):
            logger.warning(
                "Entries in field list must be indirect references -- skipping."
            )
            continue
        if field_ref.reference in refs_seen:
            raise PdfReadError("Circular reference in form tree")

        field = field_ref.get_object()
        if not isinstance(field, generic.DictionaryObject):
            logger.warning(
                "Entries in field list must be dictionary objects, not "
                f"{type(field)} -- skipping."
            )
            continue
        # /T is the field name. If not specified, we're dealing with a bare
        # widget, so skip it. (these should never occur in /Fields, but hey)
        try:
            field_name = field['/T']
        except KeyError:
            continue
        fq_name = (
            field_name
            if not parent_name
            else ("%s.%s" % (parent_name, field_name))
        )
        explicitly_requested = with_name is not None and fq_name == with_name
        child_requested = explicitly_requested or (
            with_name is not None and with_name.startswith(fq_name)
        )
        # /FT is inheritable, so go up the chain
        current_path = (field,) + parents
        for parent_field in current_path:
            try:
                field_type = parent_field['/FT']
                break
            except KeyError:
                continue
        else:
            field_type = None

        if field_type == '/Sig':
            field_value = field.get('/V')
            # "cast" to a regular string object
            filled = field_value is not None
            status_check = filled_status is None or filled == filled_status
            name_check = with_name is None or explicitly_requested
            if status_check and name_check:
                yield fq_name, field_value, field_ref
        elif explicitly_requested:
            raise SigningError(
                'Field with name %s exists but is not a signature field'
                % fq_name
            )

        # if necessary, descend into the field hierarchy
        if with_name is None or (child_requested and not explicitly_requested):
            try:
                yield from enumerate_sig_fields_in(
                    field['/Kids'],
                    parent_name=fq_name,
                    parents=current_path,
                    with_name=with_name,
                    filled_status=filled_status,
                    refs_seen=refs_seen | {field_ref.reference},
                )
            except KeyError:
                continue

def apply_sig_field_spec_properties(
    pdf_out: BasePdfFileWriter,
    sig_field: generic.DictionaryObject,
    sig_field_spec: SigFieldSpec,
):
    """
    Internal function to apply field spec properties to a newly created field.
    """

    if sig_field_spec.readable_field_name is not None:
        sig_field[pdf_name('/TU')] = generic.TextStringObject(
            sig_field_spec.readable_field_name
        )
    if sig_field_spec.seed_value_dict is not None:
        # /SV must be an indirect reference as per the spec
        sv_ref = pdf_out.add_object(
            sig_field_spec.seed_value_dict.as_pdf_object()
        )
        sig_field[pdf_name('/SV')] = sv_ref

    lock = sig_field_spec.format_lock_dictionary()
    if lock is not None:
        sig_field[pdf_name('/Lock')] = pdf_out.add_object(lock)

def ensure_sig_flags(writer: BasePdfFileWriter, lock_sig_flags: bool = True):
    """
    Ensure the SigFlags setting is present in the AcroForm dictionary.

    :param writer:
        A PDF writer.
    :param lock_sig_flags:
        Whether to flag the document as append-only.
    """
    # make sure /SigFlags is present. If not, create it
    # 3 = use append-only mode

    form = writer.root['/AcroForm']

    if lock_sig_flags:
        orig_sig_flags = form.get('/SigFlags', None)
        form['/SigFlags'] = generic.NumberObject(3)
        if orig_sig_flags != 3:
            writer.update_container(form)
    else:
        form.setdefault(pdf_name('/SigFlags'), generic.NumberObject(1))

def get_sig_field_annot(
    sig_field: generic.DictionaryObject,
) -> generic.DictionaryObject:
    """
    Internal function to get the annotation of a signature field.

    :param sig_field:
        A signature field dictionary.
    :return:
        The dictionary of the corresponding annotation.
    """
    try:
        (sig_annot,) = sig_field['/Kids']
        sig_annot = sig_annot.get_object()
    except (ValueError, TypeError):
        raise SigningError(
            "Failed to access signature field's annotation. "
            "Signature field must have exactly one child annotation, "
            "or it must be combined with its annotation."
        )
    except KeyError:
        sig_annot = sig_field
    return sig_annot

def prepare_sig_field(
    sig_field_name,
    root,
    update_writer: BasePdfFileWriter,
    existing_fields_only=False,
    **kwargs,
):
    """
    Returns a tuple of a boolean and a reference to a signature field.
    The boolean is ``True`` if the field was created, and ``False`` otherwise.

    .. danger::
        This function is internal API.
    """

    try:
        form = root['/AcroForm']

        try:
            fields = form['/Fields']
        except KeyError:
            fields = form['/Fields'] = generic.ArrayObject()

        candidates = enumerate_sig_fields_in(
            fields, with_name=sig_field_name, refs_seen=set()
        )
        sig_field_ref = None
        try:
            field_name, value, sig_field_ref = next(candidates)
            if value is not None:
                raise SigningError(
                    'Signature field with name %s appears to be filled already.'
                    % sig_field_name
                )
        except StopIteration:
            if existing_fields_only:
                raise SigningError(
                    'No empty signature field with name %s found.'
                    % sig_field_name
                )
        form_created = False
    except KeyError:
        # we have to create the form
        if existing_fields_only:
            raise SigningError('This file does not contain a form.')
        # no AcroForm present, so create one
        form = generic.DictionaryObject()
        root[pdf_name('/AcroForm')] = update_writer.add_object(form)
        fields = generic.ArrayObject()
        form[pdf_name('/Fields')] = fields
        # now we need to mark the root as updated
        update_writer.update_root()
        form_created = True
        sig_field_ref = None

    if sig_field_ref is not None:
        return False, sig_field_ref

    # no signature field exists, so create one
    # default: grab a reference to the first page
    page_ref = update_writer.find_page_for_modification(0)[0]
    sig_form_kwargs = {'include_on_page': page_ref}
    sig_form_kwargs.update(**kwargs)
    sig_field = SignatureFormField(sig_field_name, **sig_form_kwargs)
    created, sig_field_ref = _insert_or_get_field_at(
        update_writer,
        fields,
        path=sig_field_name.split('.'),
        field_obj=sig_field,
    )
    sig_field.register_widget_annotation(update_writer, sig_field_ref)

    # if a field was added to an existing form, register an extra update
    if not form_created:
        update_writer.update_container(fields)
    return True, sig_field_ref

class SignatureFormField(generic.DictionaryObject):
    def __init__(
        self,
        field_name,
        *,
        box=None,
        include_on_page=None,
        combine_annotation=True,
        invis_settings: InvisSigSettings = InvisSigSettings(),
        visible_settings: VisibleSigSettings = VisibleSigSettings(),
        annot_flags=None,
    ):
        if box is not None:
            rect = [generic.FloatObject(rd(x)) for x in box]
            invisible = not (abs(box[0] - box[2]) and abs(box[1] - box[3]))
        else:
            coord = -9999 if invis_settings.box_out_of_bounds else 0
            rect = [generic.FloatObject(coord)] * 4
            invisible = True

        super().__init__(
            {
                # Signature field properties
                pdf_name('/FT'): pdf_name('/Sig'),
                pdf_name('/T'): pdf_string(field_name),
            }
        )

        self.combine_annotation = combine_annotation
        annot_dict: generic.DictionaryObject
        if combine_annotation:
            annot_dict = self
        else:
            annot_dict = generic.DictionaryObject()

        # Annotation properties: bare minimum
        annot_dict['/Type'] = pdf_name('/Annot')
        annot_dict['/Subtype'] = pdf_name('/Widget')

        if annot_flags is None:
            # this sets the "lock" bit
            annot_flags = 0b10000000
            if invisible:
                if invis_settings.set_hidden_flag:
                    annot_flags |= 0b10
                if invis_settings.set_print_flag:
                    annot_flags |= 0b100
            else:
                if visible_settings.print_signature:
                    annot_flags |= 0b100
                if not visible_settings.scale_with_page_zoom:
                    annot_flags |= 0b1000
                if not visible_settings.rotate_with_page:
                    annot_flags |= 0b10000

        annot_dict['/F'] = generic.NumberObject(annot_flags)
        annot_dict['/Rect'] = generic.ArrayObject(rect)

        self.page_ref = include_on_page
        if include_on_page is not None:
            annot_dict['/P'] = include_on_page

        self.annot_dict = annot_dict

    def register_widget_annotation(
        self, writer: BasePdfFileWriter, sig_field_ref
    ):
        annot_dict = self.annot_dict
        if not self.combine_annotation:
            annot_ref = writer.add_object(annot_dict)
            self['/Kids'] = generic.ArrayObject([annot_ref])
        else:
            annot_ref = sig_field_ref
        writer.register_annotation(self.page_ref, annot_ref)



"""
Internal backend-agnostic utilities to help process fetched certificates, CRLs
and OCSP responses.
"""

import asyncio
import logging
import os
from typing import Awaitable, Callable, Dict, Optional, TypeVar, Union

from asn1crypto import algos, cms, core, ocsp, pem, x509

from .. import errors
from ..authority import Authority
from ..util import get_ac_extension_value

__all__ = [
    'unpack_cert_content',
    'format_ocsp_request',
    'process_ocsp_response_data',
    'queue_fetch_task',
    'crl_job_results_as_completed',
    'ocsp_job_get_earliest',
    'complete_certificate_fetch_jobs',
    'gather_aia_issuer_urls',
    'ACCEPTABLE_STRICT_CERT_CONTENT_TYPES',
    'ACCEPTABLE_CERT_PEM_ALIASES',
    'ACCEPTABLE_PKCS7_DER_ALIASES',
    'ACCEPTABLE_CERT_DER_ALIASES',
]

logger = logging.getLogger(__name__)


ACCEPTABLE_STRICT_CERT_CONTENT_TYPES = frozenset(
    [
        'application/pkix-cert',
        'application/pkcs7-mime',
        'application/x-x509-ca-cert',
        'application/x-pkcs7-certificates',
    ]
)

ACCEPTABLE_CERT_PEM_ALIASES = frozenset(
    [
        'application/x-pem-file',
        'text/plain',
        'application/octet-stream',
        'binary/octet-stream',
    ]
)

ACCEPTABLE_CERT_DER_ALIASES = frozenset(
    [
        'application/pkix-cert',
        'application/x-x509-ca-cert',
        'application/octet-stream',
        'binary/octet-stream',
    ]
)


ACCEPTABLE_PKCS7_DER_ALIASES = frozenset(
    [
        'application/pkcs7-mime',
        'application/x-pkcs7-certificates',
        'binary/octet-stream',
    ]
)


def unpack_cert_content(
    response_data: bytes,
    content_type: Optional[str],
    url: str,
    permit_pem: bool,
):
    is_pem = pem.detect(response_data)
    if (
        content_type is None or content_type in ACCEPTABLE_CERT_DER_ALIASES
    ) and not is_pem:
        # sometimes we get DER over octet-stream
        if content_type is None:
            logger.warning(
                f"Response to certificate fetch request to {url} did not "
                f"include a content type, verifying it's sequence length to "
                f"check if it is a certificate or pkcs7."
            )
        der_sequence_length = len(core.Sequence.load(response_data))
        if der_sequence_length == 2:
            yield from _unpack_der_pkcs7(response_data, url)
        elif der_sequence_length == 3:
            yield x509.Certificate.load(response_data)
    elif (content_type in ACCEPTABLE_PKCS7_DER_ALIASES) and not is_pem:
        yield from _unpack_der_pkcs7(response_data, url)
    elif permit_pem and is_pem:
        # technically, PEM is not allowed here, but of course some people don't
        # bother following the rules
        for type_name, _, data in pem.unarmor(response_data, multiple=True):
            if type_name == 'PKCS7':
                yield from _unpack_der_pkcs7(data, url)
            else:
                yield x509.Certificate.load(data)
    else:  # pragma: nocover
        raise ValueError(
            f"Failed to extract certs from {content_type} payload. "
            f"Source URL: {url}."
        )


def _unpack_der_pkcs7(pkcs7_data: bytes, pkcs7_url: str):
    content_info: cms.ContentInfo = cms.ContentInfo.load(pkcs7_data)
    cms_ct = content_info['content_type'].native
    if cms_ct != 'signed_data':
        raise ValueError(
            "Expected CMS SignedData when extracting certs from "
            "application/pkcs7-mime payload, but content type was "
            f"'{cms_ct}'. Source URL: {pkcs7_url}."
        )
    signed_data = content_info['content']
    if isinstance(signed_data['certificates'], cms.CertificateSet):
        for cert_choice in signed_data['certificates']:
            if cert_choice.name == 'certificate':
                yield cert_choice.chosen


def get_certid(
    cert: Union[x509.Certificate, cms.AttributeCertificateV2],
    authority: Authority,
    *,
    certid_hash_algo,
) -> ocsp.CertId:
    if isinstance(cert, x509.Certificate):
        serial_number = cert.serial_number
    else:
        serial_number = cert['ac_info']['serial_number'].native

    iss_name_hash = getattr(authority.name, certid_hash_algo)
    cert_id = ocsp.CertId(
        {
            'hash_algorithm': algos.DigestAlgorithm(
                {'algorithm': certid_hash_algo}
            ),
            'issuer_name_hash': iss_name_hash,
            'issuer_key_hash': getattr(authority.public_key, certid_hash_algo),
            'serial_number': serial_number,
        }
    )
    return cert_id


def format_ocsp_request(
    cert: x509.Certificate,
    authority: Authority,
    *,
    certid_hash_algo: str,
    request_nonces: bool,
):
    cert_id = get_certid(cert, authority, certid_hash_algo=certid_hash_algo)

    request = ocsp.Request(
        {
            'req_cert': cert_id,
        }
    )
    tbs_request = ocsp.TBSRequest(
        {
            'request_list': ocsp.Requests([request]),
        }
    )

    if request_nonces:
        nonce_extension = ocsp.TBSRequestExtension(
            {
                'extn_id': 'nonce',
                'critical': False,
                'extn_value': core.OctetString(os.urandom(16)),
            }
        )
        tbs_request['request_extensions'] = ocsp.TBSRequestExtensions(
            [nonce_extension]
        )

    return ocsp.OCSPRequest({'tbs_request': tbs_request})


def process_ocsp_response_data(
    response_data: bytes, *, ocsp_request: ocsp.OCSPRequest, ocsp_url: str
):
    try:
        ocsp_response = ocsp.OCSPResponse.load(response_data)
    except ValueError:
        raise errors.OCSPFetchError('Failed to parse response from OCSP server')
    status = ocsp_response['response_status'].native
    if status != 'successful':
        raise errors.OCSPValidationError(
            'OCSP server at %s returned an error. Status was \'%s\'.'
            % (ocsp_url, status)
        )

    request_nonce = ocsp_request.nonce_value
    if request_nonce:
        response_nonce = ocsp_response.nonce_value
        # if the response did not contain the nonce extension, there's no
        # point in trying to enforce it, that's the CA's problem.
        #  (I suppose we could give callers the option to mark the nonce
        #  extension as critical in the request, but that's discouraged by the
        #  specification)
        if response_nonce and (request_nonce.native != response_nonce.native):
            raise errors.OCSPValidationError(
                'Unable to verify OCSP response since the request and '
                'response nonces do not match'
            )
    return ocsp_response


T = TypeVar('T')
R = TypeVar('R')


async def queue_fetch_task(
    results: Dict[T, Union[R, Exception]],
    running_jobs: Dict[T, asyncio.Event],
    tag: T,
    async_fun: Callable[[], Awaitable[R]],
) -> Union[R, Exception]:
    # use an asyncio events to make sure that we don't attempt to re-fetch
    # the same tag while the job is running
    # Note: this uses asyncio locking, so we only transfer control
    # on 'await'.
    # We use events instead of locks because we don't care about fairness,
    # and events are easier to reason about.
    try:
        result = results[tag]
        logger.debug(
            f"Result for fetch job with tag {repr(tag)} was available in cache."
        )
        return _return_or_raise(result)
    except KeyError:
        pass
    try:
        wait_event: asyncio.Event = running_jobs[tag]
        logger.debug(f"Waiting for fetch job with tag {repr(tag)} to return...")
        # there's a fetch job running, wait for it to finish and then
        # return the result
        await wait_event.wait()
        logger.debug(
            f"Received completion signal for job with tag {repr(tag)}."
        )
        return _return_or_raise(results[tag])
    except KeyError:
        logger.debug(f"Starting new fetch job with tag {repr(tag)}...")
        # no fetch job running, run the task and store the result
        running_jobs[tag] = wait_event = asyncio.Event()
        try:
            result = await async_fun()
        except Exception as e:
            logger.debug(
                f"New fetch job with tag {repr(tag)} threw an exception: {e}"
            )
            result = e
        results[tag] = result
        logger.debug(f"New fetch job with tag {repr(tag)} returned.")
        # deregister event, notify waiters
        del running_jobs[tag]
        wait_event.set()
        return _return_or_raise(result)


def _return_or_raise(result):
    if isinstance(result, Exception):
        raise result
    return result


async def crl_job_results_as_completed(jobs):
    last_e = None
    at_least_one_success = False
    for crl_job in asyncio.as_completed(list(jobs)):
        try:
            fetched_crl = await crl_job
            yield fetched_crl
        except errors.CRLFetchError as e:
            last_e = e

    if last_e is not None and not at_least_one_success:
        raise last_e


async def cancel_all(pending_tasks):
    pending = asyncio.gather(*pending_tasks)
    pending.cancel()
    try:
        await pending
    except asyncio.CancelledError:
        pass


async def ocsp_job_get_earliest(jobs):
    queue = [asyncio.create_task(coro) for coro in jobs]
    ocsp_resp = last_e = None
    while queue:
        done, queue = await asyncio.wait(
            queue, return_when=asyncio.FIRST_COMPLETED
        )
        for ocsp_job in done:
            try:
                ocsp_resp = await ocsp_job
                break
            except errors.OCSPFetchError as e:
                last_e = e
    if ocsp_resp is not None:
        # cancel remaining fetch tasks
        await cancel_all(queue)
        return ocsp_resp
    raise last_e or errors.OCSPFetchError("No OCSP results")


def gather_aia_issuer_urls(
    cert: Union[x509.Certificate, cms.AttributeCertificateV2]
):
    if isinstance(cert, x509.Certificate):
        aia_value = cert.authority_information_access_value
    else:
        aia_value = get_ac_extension_value(cert, 'authority_information_access')
    if aia_value is None:
        return
    for entry in aia_value:
        if entry['access_method'].native == 'ca_issuers':
            location = entry['access_location']
            if location.name != 'uniform_resource_identifier':
                continue
            url = location.native
            if url.startswith('http'):
                yield url


async def complete_certificate_fetch_jobs(fetch_jobs):
    for fetch_job in asyncio.as_completed(fetch_jobs):
        try:
            certs_fetched = await fetch_job
        except errors.CertificateFetchError as e:
            logger.warning(
                f'Error during certificate fetch job, skipping... '
                f'(Error: {e})',
            )
            continue
        for cert in certs_fetched:
            yield cert



import asyncio
import contextvars
import functools

# Used as an alternative for asyncio.to_thread in python <=3.8
# Repurposed from CPython, used under the terms of the PSL
# Upstream source: https://github.com/python/cpython/blob/4b4227b907a262446b9d276c274feda2590a4e6e/Lib/asyncio/threads.py
# License: https://github.com/python/cpython/blob/4b4227b907a262446b9d276c274feda2590a4e6e/LICENSE
# Copyright (c) 2021 Python Software Foundation


async def _to_thread(func, *args, **kwargs):
    """Asynchronously run function *func* in a separate thread.
    Any *args and **kwargs supplied for this function are directly passed
    to *func*. Also, the current :class:`contextvars.Context` is propagated,
    allowing context variables from the main thread to be accessed in the
    separate thread.
    Return a coroutine that can be awaited to get the eventual result of *func*.
    """
    loop = asyncio.get_running_loop()
    ctx = contextvars.copy_context()
    func_call = functools.partial(ctx.run, func, *args, **kwargs)
    return await loop.run_in_executor(None, func_call)


to_thread = getattr(asyncio, 'to_thread', _to_thread)



# coding: utf-8

import inspect


def type_name(value):
    """
    Returns a user-readable name for the type of an object

    :param value:
        A value to get the type name of

    :return:
        A unicode string of the object's type name
    """

    if inspect.isclass(value):
        cls = value
    else:
        cls = value.__class__
    if cls.__module__ in {'builtins', '__builtin__'}:
        return cls.__name__
    return '%s.%s' % (cls.__module__, cls.__name__)



import enum
import hashlib
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, Dict, Iterator, Optional, Union

from asn1crypto import core, x509

from .aurora_certvalidator.revinfo.archival import CRLContainer, OCSPContainer

__all__ = [
    'ValidationObjectType',
    'ValidationObject',
    'POEType',
    'KnownPOE',
    'POEManager',
    'digest_for_poe',
]


@enum.unique
class ValidationObjectType(enum.Enum):
    """
    Types of validation objects recognised by ETSI TS 119 102-2.
    """

    CERTIFICATE = 'certificate'
    CRL = 'CRL'
    OCSP_RESPONSE = 'OCSPResponse'
    TIMESTAMP = 'timestamp'
    EVIDENCE_RECORD = 'evidencerecord'
    PUBLIC_KEY = 'publicKey'
    SIGNED_DATA = 'signedData'
    OTHER = 'other'

    def urn(self):
        return f'urn:etsi:019102:validationObject:{self.value}'


KnownObjectType = Union[bytes, CRLContainer, OCSPContainer, x509.Certificate]


def guess_validation_object_type(
    thing: object,
) -> Optional[ValidationObjectType]:
    if isinstance(thing, CRLContainer):
        return ValidationObjectType.CRL
    elif isinstance(thing, OCSPContainer):
        return ValidationObjectType.OCSP_RESPONSE
    elif isinstance(thing, x509.Certificate):
        return ValidationObjectType.CERTIFICATE
    return None


@dataclass(frozen=True)
class ValidationObject:
    """
    A validation object used in the course of a validation operation
    for which proofs of existence can potentially be gathered.
    """

    object_type: ValidationObjectType
    """
    The type of validation object.
    """

    value: Any
    """
    The actual object.

    Currently, the following types are supported explicitly.
    Others must currently be supplied as :class:`bytes`.

     - :class:`.CRLContainer`: :attr:`.ValidationObjectType.CRL`
     - :class:`.OCSPContainer`: :attr:`.ValidationObjectType.OCSP_RESPONSE`
     - :class:`x509.Certificate`: :attr:`.ValidationObjectType.CERTIFICATE`
    """


@enum.unique
class POEType(enum.Enum):
    PROVIDED = 'provided'
    VALIDATION = 'validation'
    POLICY = 'policy'

    @property
    def urn(self) -> str:
        return f'urn:etsi:019102:poetype:{self.value}'


@dataclass(frozen=True)
class KnownPOE:
    poe_type: POEType
    digest: bytes
    poe_time: datetime
    validation_object: Optional[ValidationObject] = None


def digest_for_poe(data: bytes) -> bytes:
    return hashlib.sha256(data).digest()


class POEManager:
    """
    Class to manage proof-of-existence (POE) claims.

    :param current_dt_override:
        Override the current time.
    """

    def __init__(self, current_dt_override: Optional[datetime] = None):
        self._poes: Dict[bytes, KnownPOE] = {}
        self._current_dt_override = current_dt_override

    def register(
        self,
        data: KnownObjectType,
        poe_type: POEType,
        dt: Optional[datetime] = None,
    ) -> KnownPOE:
        """
        Register a new POE claim if no POE for an earlier time is available.

        :param data:
            Data to register a POE claim for.
        :param poe_type:
            The type of POE.
        :param dt:
            The POE time to register. If ``None``, assume the current time.
        :return:
            The oldest POE datetime available.
        """
        if isinstance(data, bytes):
            b_data = data
        elif isinstance(data, core.Asn1Value):
            b_data = data.dump()
        elif isinstance(data, CRLContainer):
            b_data = data.crl_data.dump()
        elif isinstance(data, OCSPContainer):
            b_data = data.ocsp_response_data.dump()
        else:
            raise NotImplementedError
        digest = digest_for_poe(b_data)

        dt = dt or self._current_dt_override or datetime.now(timezone.utc)
        vo_type = guess_validation_object_type(data)
        vo = None
        if vo_type:
            vo = ValidationObject(object_type=vo_type, value=data)
        return self.register_known_poe(
            KnownPOE(
                poe_type=poe_type,
                digest=digest,
                poe_time=dt,
                validation_object=vo,
            )
        )

    def register_by_digest(
        self,
        digest: bytes,
        poe_type: POEType,
        dt: Optional[datetime] = None,
    ) -> KnownPOE:
        """
        Register a new POE claim if no POE for an earlier time is available.

        :param digest:
            SHA-256 digest of the data to register a POE claim for.
        :param dt:
            The POE time to register. If ``None``, assume the current time.
        :param poe_type:
            The type of POE.
        :return:
            The oldest POE datetime available.
        """
        dt = dt or self._current_dt_override or datetime.now(timezone.utc)
        return self.register_known_poe(
            KnownPOE(
                poe_type=poe_type,
                digest=digest,
                poe_time=dt,
                validation_object=None,
            )
        )

    def register_known_poe(self, known_poe: KnownPOE) -> KnownPOE:
        """
        Register a new POE claim if no POE for an earlier time is available.

        :param known_poe:
            The POE object to register.
        :return:
            The oldest POE for the given digest.
        """
        dt = known_poe.poe_time
        digest = known_poe.digest
        try:
            cur_poe = self._poes[digest]
            if cur_poe.poe_time <= dt:
                return cur_poe
        except KeyError:
            pass
        self._poes[digest] = known_poe
        return known_poe

    def __iter__(self) -> Iterator[KnownPOE]:
        """
        Iterate over the current earliest known POE for all items currently
        being managed.

        Returns an iterator with :class:`KnownPOE` objects.
        """
        return iter(self._poes.values())

    def __getitem__(self, item: KnownObjectType) -> datetime:
        """
        Return the earliest available POE for an item.

        .. note::
            This is a wrapper around :meth:`register` with `dt=None`, and hence
            will register the current time as the POE time for the given item.
            This side effect is intentional.

        :param item:
            Item to get the current POE time for.
        :return:
            A datetime object representing the earliest available POE for the
            item.
        """
        return self.register(
            item, poe_type=POEType.VALIDATION, dt=None
        ).poe_time

    def __ior__(self, other):
        """
        Combine data in another POE manager with the POEs managed by this
        instance.
        """
        if not isinstance(other, POEManager):
            raise TypeError
        for poe in iter(other):
            self.register_known_poe(poe)

    def __copy__(self):
        new_instance = POEManager(current_dt_override=self._current_dt_override)
        new_instance._poes = dict(self._poes)
        return new_instance



import logging
from typing import Iterable, Union

import requests
from asn1crypto import cms, crl, pem, x509

from ... import errors
from ...util import get_relevant_crl_dps, issuer_serial
from ..api import CRLFetcher
from ..common_utils import crl_job_results_as_completed
from .util import RequestsFetcherMixin

logger = logging.getLogger(__name__)


class RequestsCRLFetcher(CRLFetcher, RequestsFetcherMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._by_cert = {}

    async def fetch(
        self,
        cert: Union[x509.Certificate, cms.AttributeCertificateV2],
        *,
        use_deltas=True,
    ):
        iss_serial = issuer_serial(cert)
        try:
            return self._by_cert[iss_serial]
        except KeyError:
            pass

        results = []
        async for fetched_crl in self._fetch(cert, use_deltas=use_deltas):
            results.append(fetched_crl)
        self._by_cert[iss_serial] = results
        return results

    async def _fetch_single(self, url):
        async def task():
            logger.info(f"Requesting CRL from {url}...")
            try:
                response = await self._get(
                    url, acceptable_content_types=('application/pkix-crl',)
                )
                data = response.content
                if pem.detect(data):
                    _, _, data = pem.unarmor(data)
                return crl.CertificateList.load(data)
            except (ValueError, requests.RequestException) as e:
                raise errors.CRLFetchError(
                    f"Failure to fetch CRL from URL {url}"
                ) from e

        return await self._perform_fetch(url, task)

    async def _fetch(self, cert: x509.Certificate, *, use_deltas):
        sources = get_relevant_crl_dps(cert, use_deltas=use_deltas)

        def _fetch_jobs():
            for distribution_point in sources:
                url = distribution_point.url
                # Only fetch CRLs over http
                #  (or https, but that doesn't really happen all that often)
                # In particular, don't attempt to grab CRLs over LDAP
                if url.startswith('http'):
                    yield self._fetch_single(url)

        async for result in crl_job_results_as_completed(_fetch_jobs()):
            yield result

    def fetched_crls(self) -> Iterable[crl.CertificateList]:
        return {crl_ for crl_ in self.get_results()}

    def fetched_crls_for_cert(self, cert) -> Iterable[crl.CertificateList]:
        return self._by_cert[issuer_serial(cert)]




import hashlib
import logging
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime
from typing import Dict, List, Optional, Set, Tuple, Union

from asn1crypto import cms, crl, x509
from cryptography.exceptions import InvalidSignature

from .aurora_certvalidator._state import ValProcState
from .aurora_certvalidator.authority import Authority, AuthorityWithCert
from .aurora_certvalidator.context import ValidationContext
from .aurora_certvalidator.errors import (
    CertificateFetchError,
    CRLNoMatchesError,
    CRLValidationError,
    PathValidationError,
    PSSParameterMismatch,
    RevokedError,
    CRLValidationIndeterminateError
)
from .aurora_certvalidator.ltv.types import ValidationTimingParams
from .aurora_certvalidator.path import ValidationPath
from .aurora_certvalidator.policy_decl import CertRevTrustPolicy
from .aurora_certvalidator.registry import CertificateRegistry
from .aurora_certvalidator.revinfo._err_gather import Errors
from .aurora_certvalidator.revinfo.archival import (
    CRLContainer,
    RevinfoUsabilityRating,
)
from .aurora_certvalidator.revinfo.constants import (
    KNOWN_CRL_ENTRY_EXTENSIONS,
    KNOWN_CRL_EXTENSIONS,
    VALID_REVOCATION_REASONS,
)
from .aurora_certvalidator.revinfo.manager import RevinfoManager
from .aurora_certvalidator.util import (
    ConsList,
    get_ac_extension_value,
    get_issuer_dn,
    validate_sig,
)

logger = logging.getLogger(__name__)

@dataclass
class _CRLErrs(Errors):
    issuer_failures: int = 0

@dataclass
class _CRLIssuerSearchErrs:
    candidate_issuers: int
    candidates_skipped: int = 0
    signatures_failed: int = 0
    unauthorized_certs: int = 0
    path_building_failures: int = 0
    explicit_errors: List[CRLValidationError] = field(default_factory=list)

    def get_exc(self):
        plural = self.candidate_issuers > 1
        if (
            not self.candidate_issuers
            or self.candidates_skipped == self.candidate_issuers
        ):
            return CRLNoMatchesError()
        elif self.signatures_failed == self.candidate_issuers:
            return CRLValidationError('CRL signature could not be verified')
        elif self.unauthorized_certs == self.candidate_issuers:
            return CRLValidationError(
                'The CRL issuers that were identified are not authorized '
                'to sign CRLs'
                if plural
                else 'The CRL issuer that was identified is '
                'not authorized to sign CRLs'
            )
        elif self.path_building_failures == self.candidate_issuers:
            return CRLValidationError(
                'The chain of trust for the CRL issuers that were identified '
                'could not be determined'
                if plural
                else 'The chain of trust for the CRL issuer that was identified '
                'could not be determined'
            )
        elif self.explicit_errors and len(self.explicit_errors) == 1:
            # if there's only one error, throw it
            return self.explicit_errors[0]
        else:
            msg = 'Unable to determine CRL trust status. '
            msg += '; '.join(str(e) for e in self.explicit_errors)
            return CRLValidationError(msg)
        
async def _classify_relevant_crls(
    revinfo_manager: RevinfoManager,
    cert: x509.Certificate,
    errs: _CRLErrs,
    control_time: Optional[datetime] = None,
):
    # NOTE: the control_time parameter is only used in the time sliding
    # algorithm code path for AdES validation

    certificate_lists = await revinfo_manager.async_retrieve_crls(cert)
    poe_manager = revinfo_manager.poe_manager

    complete_lists_by_issuer = defaultdict(list)
    delta_lists_by_issuer = defaultdict(list)
    for certificate_list_cont in certificate_lists:
        certificate_list = certificate_list_cont.crl_data
        if control_time is not None:
            issued = certificate_list_cont.issuance_date
            if (
                issued is None
                or issued > control_time
                or poe_manager[certificate_list_cont] > control_time
            ):
                # We don't care about stuff issued after control_time
                # or without the right POE
                continue
        try:
            issuer_hashable = certificate_list.issuer.hashable
            if certificate_list.delta_crl_indicator_value is None:
                complete_lists_by_issuer[issuer_hashable].append(
                    certificate_list_cont
                )
            else:
                delta_lists_by_issuer[issuer_hashable].append(
                    certificate_list_cont
                )
        except ValueError as e:
            msg = "Generic processing error while classifying CRL."
            logging.debug(msg, exc_info=e)
            errs.append(msg, certificate_list)
    return complete_lists_by_issuer, delta_lists_by_issuer

def _get_crl_authority_name(
    certificate_list_cont: CRLContainer,
    cert_issuer_name: x509.Name,
    certificate_registry: CertificateRegistry,
    errs: _CRLErrs,
) -> Tuple[bool, x509.Name]:
    """
    Figure out the name of the entity on behalf of which the CRL was issued.
    """

    certificate_list = certificate_list_cont.crl_data

    crl_idp: crl.IssuingDistributionPoint = (
        certificate_list.issuing_distribution_point_value
    )
    is_indirect = bool(crl_idp and crl_idp['indirect_crl'].native)
    if not is_indirect:
        crl_authority_name = certificate_list.issuer
    else:
        crl_idp_name = crl_idp['distribution_point']
        if crl_idp_name:
            if crl_idp_name.name == 'full_name':
                crl_authority_name = crl_idp_name.chosen[0].chosen
            else:
                crl_authority_name = cert_issuer_name.copy().chosen.append(
                    crl_idp_name.chosen
                )
        elif certificate_list.authority_key_identifier:
            tmp_crl_issuer = certificate_registry.retrieve_by_key_identifier(
                certificate_list.authority_key_identifier
            )
            crl_authority_name = tmp_crl_issuer.subject
        else:
            errs.append(
                'CRL is marked as an indirect CRL, but provides no '
                'mechanism for locating the CRL issuer certificate',
                certificate_list_cont,
            )
            raise LookupError
    return is_indirect, crl_authority_name

async def _find_candidate_crl_issuer_certs(
    crl_authority_name: x509.Name,
    certificate_list: crl.CertificateList,
    *,
    cert_issuer_auth: Authority,
    cert_registry: CertificateRegistry,
) -> List[x509.Certificate]:
    # first, look for certs issued to the issuer named as the entity
    # that signed the CRL.
    # In both cases, we prioritise the next-level issuer in the main path
    # if it matches the criteria.
    delegated_issuer = certificate_list.issuer
    cert_issuer_cert = None
    if isinstance(cert_issuer_auth, AuthorityWithCert):
        cert_issuer_cert = cert_issuer_auth.certificate
    candidates = cert_registry.retrieve_by_name(
        delegated_issuer, cert_issuer_cert
    )
    if not candidates and crl_authority_name != certificate_list.issuer:
        # next, look in the cache for certs issued to the entity named
        # in the issuing distribution point (i.e. the issuing authority)
        candidates = cert_registry.retrieve_by_name(
            crl_authority_name, cert_issuer_cert
        )
    if not candidates and cert_registry.fetcher is not None:
        candidates = []
        valid_names = {crl_authority_name, delegated_issuer}
        # Try to download certificates from URLs in the AIA extension,
        # if there is one
        async for cert in cert_registry.fetcher.fetch_crl_issuers(
            certificate_list
        ):
            # filter by name
            if cert.subject in valid_names:
                candidates.insert(0, cert)
    return candidates

def _verify_crl_signature(certificate_list, public_key):
    """
    Verifies the digital signature on an asn1crypto.crl.CertificateList object

    :param certificate_list:
        An asn1crypto.crl.CertificateList object

    :raises:
        aurora_certvalidator.errors.CRLValidationError - when the signature is
        invalid or uses an unsupported algorithm
    """

    signature_algo = certificate_list['signature_algorithm'].signature_algo
    hash_algo = certificate_list['signature_algorithm'].hash_algo

    try:
        validate_sig(
            signature=certificate_list['signature'].native,
            signed_data=certificate_list['tbs_cert_list'].dump(),
            public_key_info=public_key,
            sig_algo=signature_algo,
            hash_algo=hash_algo,
            parameters=certificate_list['signature_algorithm']['parameters'],
        )
    except PSSParameterMismatch as e:
        raise CRLValidationError(
            'Invalid signature parameters on CertificateList'
        ) from e
    except InvalidSignature:
        raise CRLValidationError(
            'Unable to verify the signature of the CertificateList'
        )
    
async def _find_candidate_crl_paths(
    crl_authority_name: x509.Name,
    certificate_list: crl.CertificateList,
    *,
    cert: Union[x509.Certificate, cms.AttributeCertificateV2],
    cert_issuer_auth: Authority,
    cert_path: ValidationPath,
    certificate_registry: CertificateRegistry,
    is_indirect: bool,
    proc_state: ValProcState,
) -> Tuple[List[ValidationPath], _CRLIssuerSearchErrs]:
    cert_sha256 = hashlib.sha256(cert.dump()).digest()

    candidate_crl_issuers = await _find_candidate_crl_issuer_certs(
        crl_authority_name,
        certificate_list,
        cert_issuer_auth=cert_issuer_auth,
        cert_registry=certificate_registry,
    )
    cert_issuer_name = cert_issuer_auth.name

    errs = _CRLIssuerSearchErrs(candidate_issuers=len(candidate_crl_issuers))
    candidate_paths = []
    for candidate_crl_issuer in candidate_crl_issuers:
        direct_issuer = candidate_crl_issuer.subject == cert_issuer_name

        # In some cases an indirect CRL issuer is a certificate issued
        # by the certificate issuer. However, we need to ensure that
        # the candidate CRL issuer is not the certificate being checked,
        # otherwise we may be checking an incorrect CRL and produce
        # incorrect results.
        indirect_issuer = (
            candidate_crl_issuer.issuer == cert_issuer_name
            and candidate_crl_issuer.sha256 != cert_sha256
        )

        if not direct_issuer and not indirect_issuer and not is_indirect:
            errs.candidates_skipped += 1
            continue

        key_usage_value = candidate_crl_issuer.key_usage_value
        if key_usage_value and 'crl_sign' not in key_usage_value.native:
            errs.unauthorized_certs += 1
            continue

        try:
            # Step g
            # NOTE: Theoretically this can only be done after full X.509
            # path validation (step f), but that only matters for DSA key
            # inheritance which we don't support anyhow when doing revocation
            # checks.
            _verify_crl_signature(
                certificate_list, candidate_crl_issuer.public_key
            )
        except CRLValidationError:
            errs.signatures_failed += 1
            continue

        cand_path = proc_state.check_path_verif_recursion(candidate_crl_issuer)
        if not cand_path:
            try:
                cand_path = cert_path.truncate_to_issuer_and_append(
                    candidate_crl_issuer
                )
            except LookupError:
                errs.path_building_failures += 1
                continue
        candidate_paths.append(cand_path)
    return candidate_paths, errs

async def _validate_crl_issuer_path(
    *,
    candidate_crl_issuer_path: ValidationPath,
    validation_context: ValidationContext,
    issuing_authority_identical: bool,
    proc_state: ValProcState,
):
    # If we have a validation cached (from before, or because the CRL issuer
    #  appears further up in the path) use it.
    # This is not just for efficiency, it also makes for clearer errors when
    #  validation fails due to revocation info issues further up in the path
    if validation_context.check_validation(candidate_crl_issuer_path.last):
        return
    try:
        temp_override = proc_state.ee_name_override
        if not issuing_authority_identical:
            temp_override = (
                proc_state.describe_cert(never_def=True) + ' CRL issuer'
            )
        from .aurora_certvalidator.validate import intl_validate_path

        new_stack = proc_state.cert_path_stack.cons(candidate_crl_issuer_path)
        await intl_validate_path(
            validation_context,
            candidate_crl_issuer_path,
            proc_state=ValProcState(
                ee_name_override=temp_override, cert_path_stack=new_stack
            ),
        )

    except PathValidationError as e:
        iss_cert = candidate_crl_issuer_path.last
        logger.warning(
            f"Path for CRL issuer {iss_cert.subject.human_friendly} could not "
            f"be validated.",
            exc_info=e,
        )
        raise CRLValidationError(
            f'The CRL issuer certificate path could not be validated. {e}'
        )

async def _find_crl_issuer(
    crl_authority_name: x509.Name,
    certificate_list: crl.CertificateList,
    *,
    cert: Union[x509.Certificate, cms.AttributeCertificateV2],
    cert_issuer_auth: Authority,
    cert_path: ValidationPath,
    validation_context: ValidationContext,
    is_indirect: bool,
    proc_state: ValProcState,
) -> ValidationPath:
    candidate_paths, errs = await _find_candidate_crl_paths(
        crl_authority_name,
        certificate_list,
        cert=cert,
        cert_issuer_auth=cert_issuer_auth,
        cert_path=cert_path,
        certificate_registry=validation_context.certificate_registry,
        is_indirect=is_indirect,
        proc_state=proc_state,
    )

    for candidate_crl_issuer_path in candidate_paths:
        candidate_crl_issuer = candidate_crl_issuer_path.last

        # Skip path validation step if we're recursing
        #  (necessary to process CRLs that have their own certificate in-scope,
        #   which is questionable practice, but PKITS has a test case for this
        #   specific wrinkle, and it's not contradicted by anything in RFC 5280,
        #   so it's probably allowed in theory)
        if proc_state.check_path_verif_recursion(candidate_crl_issuer):
            validation_context.revinfo_manager.record_crl_issuer(
                certificate_list, candidate_crl_issuer
            )
            return candidate_crl_issuer_path
        # Step f
        # Note: this is not the same as .truncate_to() if
        # candidate_crl_issuer doesn't appear in the path!
        candidate_crl_issuer_path = cert_path.truncate_to_issuer_and_append(
            candidate_crl_issuer
        )
        try:
            # This check needs to know not only whether the names agree,
            # but also whether the keys are the same, in order to yield
            # the correct error message on failure.
            # (Scenario: CA with separate keys for CRL signing and for
            # certificate issuance, but with the same name on both certs)
            issuing_authority_identical = not is_indirect and (
                cert_issuer_auth is not None
                and cert_issuer_auth.public_key.dump()
                == candidate_crl_issuer.public_key.dump()
            )
            await _validate_crl_issuer_path(
                candidate_crl_issuer_path=candidate_crl_issuer_path,
                validation_context=validation_context,
                issuing_authority_identical=issuing_authority_identical,
                proc_state=proc_state,
            )
            validation_context.revinfo_manager.record_crl_issuer(
                certificate_list, candidate_crl_issuer
            )
            return candidate_crl_issuer_path
        except CRLValidationError as e:
            errs.explicit_errors.append(e)
            continue
    raise errs.get_exc()

def _match_dps_idp_names(
    crl_idp: crl.IssuingDistributionPoint,
    crl_dps: Optional[x509.CRLDistributionPoints],
    crl_issuer: x509.Certificate,
    crl_authority_name: x509.Name,
) -> bool:
    # Step b 2 i
    has_idp_name = False
    has_dp_name = False
    idp_dp_match = False

    idp_general_names = []
    idp_dp_name = crl_idp['distribution_point']
    if idp_dp_name:
        has_idp_name = True
        if idp_dp_name.name == 'full_name':
            for general_name in idp_dp_name.chosen:
                idp_general_names.append(general_name)
        else:
            inner_extended_issuer_name = crl_issuer.subject.copy()
            inner_extended_issuer_name.chosen.append(idp_dp_name.chosen.untag())
            idp_general_names.append(
                x509.GeneralName(
                    name='directory_name', value=inner_extended_issuer_name
                )
            )

    if crl_dps:
        for dp in crl_dps:
            if idp_dp_match:
                break
            dp_name = dp['distribution_point']
            if dp_name:
                has_dp_name = True
                if dp_name.name == 'full_name':
                    for general_name in dp_name.chosen:
                        if general_name in idp_general_names:
                            idp_dp_match = True
                            break
                else:
                    inner_extended_issuer_name = crl_issuer.subject.copy()
                    inner_extended_issuer_name.chosen.append(
                        dp_name.chosen.untag()
                    )
                    dp_extended_issuer_name = x509.GeneralName(
                        name='directory_name', value=inner_extended_issuer_name
                    )

                    if dp_extended_issuer_name in idp_general_names:
                        idp_dp_match = True

            elif dp['crl_issuer']:
                has_dp_name = True
                for dp_crl_authority_name in dp['crl_issuer']:
                    if dp_crl_authority_name in idp_general_names:
                        idp_dp_match = True
                        break
    else:
        # If there is no DP, we consider the CRL issuer name to be it
        has_dp_name = True
        general_name = x509.GeneralName(
            name='directory_name', value=crl_authority_name
        )
        if general_name in idp_general_names:
            idp_dp_match = True

    return idp_dp_match or not has_idp_name or not has_dp_name

def _handle_crl_idp_ext_constraints(
    cert: x509.Certificate,
    certificate_list: crl.CertificateList,
    crl_issuer: x509.Certificate,
    crl_idp: crl.IssuingDistributionPoint,
    crl_authority_name: x509.Name,
    errs: _CRLErrs,
) -> bool:
    match = _match_dps_idp_names(
        crl_idp=crl_idp,
        crl_dps=cert.crl_distribution_points_value,
        crl_issuer=crl_issuer,
        crl_authority_name=crl_authority_name,
    )
    if not match:
        errs.append(
            "The CRL issuing distribution point extension does not "
            "share any names with the certificate CRL distribution "
            "point extension",
            certificate_list,
        )
        errs.issuer_failures += 1
        return False

    # Step b 2 ii
    if crl_idp['only_contains_user_certs'].native:
        if (
            cert.basic_constraints_value
            and cert.basic_constraints_value['ca'].native
        ):
            errs.append(
                "CRL only contains end-entity certificates and "
                "certificate is a CA certificate",
                certificate_list,
            )
            return False

    # Step b 2 iii
    if crl_idp['only_contains_ca_certs'].native:
        if (
            not cert.basic_constraints_value
            or cert.basic_constraints_value['ca'].native is False
        ):
            errs.append(
                "CRL only contains CA certificates and certificate "
                "is an end-entity certificate",
                certificate_list,
            )
            return False

    # Step b 2 iv
    if crl_idp['only_contains_attribute_certs'].native:
        errs.append(
            'CRL only contains attribute certificates', certificate_list
        )
        return False

    return True

def _handle_attr_cert_crl_idp_ext_constraints(
    certificate_list: crl.CertificateList,
    crl_dps: Optional[x509.CRLDistributionPoints],
    crl_issuer: x509.Certificate,
    crl_idp: crl.IssuingDistributionPoint,
    crl_authority_name: x509.Name,
    errs: _CRLErrs,
) -> bool:
    match = _match_dps_idp_names(
        crl_idp=crl_idp,
        crl_dps=crl_dps,
        crl_issuer=crl_issuer,
        crl_authority_name=crl_authority_name,
    )
    if not match:
        errs.append(
            "The CRL issuing distribution point extension does not "
            "share any names with the attribute certificate's "
            "CRL distribution point extension",
            certificate_list,
        )
        errs.issuer_failures += 1
        return False

    # Step b 2 ii
    pkc_only = (
        crl_idp['only_contains_user_certs'].native
        or crl_idp['only_contains_ca_certs'].native
    )
    if pkc_only:
        errs.append(
            "CRL only contains public-key certificates, but "
            "certificate is an attribute certificate",
            certificate_list,
        )
        return False

    return True

def _verify_no_unknown_critical_extensions(
    certificate_list_cont: CRLContainer, errs: _CRLErrs, is_delta: bool
):
    extensions = certificate_list_cont.crl_data.critical_extensions
    if extensions - KNOWN_CRL_EXTENSIONS:
        errs.append(
            f'One or more unrecognized critical extensions are present in '
            f'the {"delta CRL" if is_delta else "CRL"}',
            certificate_list_cont,
        )
        return False
    return True

def _get_crl_scope_assuming_authority(
    crl_issuer: x509.Certificate,
    cert: Union[x509.Certificate, cms.AttributeCertificateV2],
    certificate_list_cont: CRLContainer,
    is_indirect: bool,
    errs: _CRLErrs,
) -> Optional[Set[str]]:
    certificate_list = certificate_list_cont.crl_data
    crl_idp: crl.IssuingDistributionPoint = (
        certificate_list.issuing_distribution_point_value
    )

    is_pkc = isinstance(cert, x509.Certificate)
    # Step b 1
    has_dp_crl_issuer = False
    dp_match = False

    if is_pkc:
        crl_dps = cert.crl_distribution_points_value
    else:
        crl_dps = get_ac_extension_value(cert, 'crl_distribution_points')
    if crl_dps:
        crl_issuer_general_name = x509.GeneralName(
            name='directory_name', value=crl_issuer.subject
        )
        for dp in crl_dps:
            if dp['crl_issuer']:
                has_dp_crl_issuer = True
                if crl_issuer_general_name in dp['crl_issuer']:
                    dp_match = True

    crl_authority_name = crl_issuer.subject
    cert_issuer_name = get_issuer_dn(cert)
    same_issuer = crl_authority_name == cert_issuer_name
    indirect_match = has_dp_crl_issuer and dp_match and is_indirect
    missing_idp = has_dp_crl_issuer and (not dp_match or not is_indirect)
    indirect_crl_issuer = crl_issuer.issuer == cert_issuer_name

    if (
        not same_issuer and not indirect_match and not indirect_crl_issuer
    ) or missing_idp:
        errs.issuer_failures += 1
        return None

    # Step b 2

    if crl_idp is not None:
        if is_pkc:
            crl_idp_match = _handle_crl_idp_ext_constraints(
                cert=cert,
                certificate_list=certificate_list,
                crl_issuer=crl_issuer,
                crl_idp=crl_idp,
                crl_authority_name=crl_authority_name,
                errs=errs,
            )
        else:
            crl_idp_match = _handle_attr_cert_crl_idp_ext_constraints(
                crl_dps=crl_dps,
                certificate_list=certificate_list,
                crl_issuer=crl_issuer,
                crl_idp=crl_idp,
                crl_authority_name=crl_authority_name,
                errs=errs,
            )
        # error reporting is taken care of in the delegated method
        if not crl_idp_match:
            return None

    # Step d
    idp_reasons = None

    if crl_idp and crl_idp['only_some_reasons'].native is not None:
        idp_reasons = crl_idp['only_some_reasons'].native

    reason_keys = None
    if idp_reasons:
        reason_keys = idp_reasons

    if reason_keys is None:
        interim_reasons = VALID_REVOCATION_REASONS.copy()
    else:
        interim_reasons = reason_keys

    # Step e
    # We don't skip a CRL if it only contains reasons already checked since
    # a certificate issuer can self-issue a new cert that is used for CRLs

    if not _verify_no_unknown_critical_extensions(
        certificate_list_cont, errs, is_delta=False
    ):
        return None

    return interim_reasons

def _find_matching_delta_crl(
    delta_lists: List[CRLContainer],
    crl_authority_name: x509.Name,
    crl_idp: crl.IssuingDistributionPoint,
    parent_crl_aki: Optional[bytes],
) -> Optional[CRLContainer]:
    for candidate_delta_cl_cont in delta_lists:
        candidate_delta_cl = candidate_delta_cl_cont.crl_data
        # Step c 1
        if candidate_delta_cl.issuer != crl_authority_name:
            continue

        # Step c 2
        delta_crl_idp = candidate_delta_cl.issuing_distribution_point_value
        if (crl_idp is None and delta_crl_idp is not None) or (
            crl_idp is not None and delta_crl_idp is None
        ):
            continue

        if crl_idp is not None and crl_idp.native != delta_crl_idp.native:
            continue

        # Step c 3
        if parent_crl_aki != candidate_delta_cl.authority_key_identifier:
            continue

        return candidate_delta_cl_cont
    return None

def _check_crl_freshness(
    certificate_list_cont: CRLContainer,
    revinfo_policy: CertRevTrustPolicy,
    timing_params: ValidationTimingParams,
    errs: _CRLErrs,
    is_delta: bool,
):
    freshness_result = certificate_list_cont.usable_at(
        policy=revinfo_policy,
        timing_params=timing_params,
    )
    prefix = "Delta CRL" if is_delta else "CRL"
    rating = freshness_result.rating
    if rating != RevinfoUsabilityRating.OK:
        if rating == RevinfoUsabilityRating.STALE:
            msg = f'{prefix} is not recent enough'
            errs.update_stale(freshness_result.last_usable_at)
        elif rating == RevinfoUsabilityRating.TOO_NEW:
            msg = f'{prefix} is too recent'
        else:
            msg = f'{prefix} freshness could not be established'
        errs.append(msg, certificate_list_cont, is_freshness_failure=True)
        return False
    return True

def _maybe_get_delta_crl(
    certificate_list: crl.CertificateList,
    crl_issuer: x509.Certificate,
    delta_lists_by_issuer: Dict[str, List[CRLContainer]],
    errs: _CRLErrs,
    timing_params: Optional[ValidationTimingParams] = None,
    policy: Optional[CertRevTrustPolicy] = None,
) -> Optional[CRLContainer]:
    if (
        not certificate_list.freshest_crl_value
        or len(certificate_list.freshest_crl_value) == 0
    ):
        # nothing to do, return
        return None

    crl_authority_name = crl_issuer.subject
    crl_idp: crl.IssuingDistributionPoint = (
        certificate_list.issuing_distribution_point_value
    )

    candidate_delta_lists = delta_lists_by_issuer.get(
        crl_authority_name.hashable, []
    )
    delta_certificate_list_cont = _find_matching_delta_crl(
        delta_lists=candidate_delta_lists,
        crl_authority_name=crl_authority_name,
        crl_idp=crl_idp,
        parent_crl_aki=certificate_list.authority_key_identifier,
    )
    if not delta_certificate_list_cont:
        return None

    delta_certificate_list = delta_certificate_list_cont.crl_data

    if not _verify_no_unknown_critical_extensions(
        delta_certificate_list_cont, errs, is_delta=True
    ):
        return None

    # Step h
    try:
        _verify_crl_signature(delta_certificate_list, crl_issuer.public_key)
    except CRLValidationError:
        errs.append(
            'Delta CRL signature could not be verified',
            delta_certificate_list_cont,
        )
        return None

    if policy and timing_params:
        if _check_crl_freshness(
            delta_certificate_list_cont,
            policy,
            timing_params,
            errs,
            is_delta=True,
        ):
            return delta_certificate_list_cont
    return None

def find_cert_in_list(
    cert: Union[x509.Certificate, cms.AttributeCertificateV2],
    cert_issuer_name: x509.Name,
    certificate_list: crl.CertificateList,
    crl_authority_name: x509.Name,
):
    """
    Looks for a cert in the list of revoked certificates

    :param cert:
        An asn1crypto.x509.Certificate object of the cert being checked,
        or an asn1crypto.cms.AttributeCertificateV2 object in the case
        of an attribute certificate.

    :param cert_issuer_name:
        The certificate issuer's distinguished name

    :param certificate_list:
        An ans1crypto.crl.CertificateList object to look in for the cert

    :param crl_authority_name:
        The distinguished name of the default authority for which the CRL issues
        certificates.

    :return:
        A tuple of (None, None) if not present, otherwise a tuple of
        (asn1crypto.x509.Time object, asn1crypto.crl.CRLReason object)
        representing the date/time the object was revoked and why
    """

    revoked_certificates = certificate_list['tbs_cert_list'][
        'revoked_certificates'
    ]

    if isinstance(cert, x509.Certificate):
        cert_serial = cert.serial_number
    else:
        cert_serial = cert['ac_info']['serial_number'].native

    last_issuer_name = crl_authority_name
    for revoked_cert in revoked_certificates:
        # If any unknown critical extensions, the entry can not be used
        if revoked_cert.critical_extensions - KNOWN_CRL_ENTRY_EXTENSIONS:
            raise NotImplementedError()

        if (
            revoked_cert.issuer_name
            and revoked_cert.issuer_name != last_issuer_name
        ):
            last_issuer_name = revoked_cert.issuer_name
        if last_issuer_name != cert_issuer_name:
            continue

        if revoked_cert['user_certificate'].native != cert_serial:
            continue

        if not revoked_cert.crl_reason_value:
            crl_reason = crl.CRLReason('unspecified')
        else:
            crl_reason = revoked_cert.crl_reason_value

        return revoked_cert['revocation_date'].native, crl_reason

    return None, None

def _check_cert_on_crl_and_delta(
    crl_issuer: x509.Certificate,
    cert: Union[x509.Certificate, cms.AttributeCertificateV2],
    certificate_list_cont: CRLContainer,
    delta_certificate_list_cont: Optional[CRLContainer],
    errs: _CRLErrs,
):
    certificate_list = certificate_list_cont.crl_data
    # Step i
    revoked_reason = None
    revoked_date = None

    cert_issuer_name = get_issuer_dn(cert)

    if delta_certificate_list_cont:
        delta_certificate_list = delta_certificate_list_cont.crl_data
        try:
            revoked_date, revoked_reason = find_cert_in_list(
                cert,
                cert_issuer_name,
                delta_certificate_list,
                crl_issuer.subject,
            )
        except NotImplementedError:
            errs.append(
                'One or more unrecognized critical extensions are present in '
                'the CRL entry for the certificate',
                delta_certificate_list_cont,
            )
            raise

    # Step j
    if revoked_reason is None:
        try:
            revoked_date, revoked_reason = find_cert_in_list(
                cert, cert_issuer_name, certificate_list, crl_issuer.subject
            )
        except NotImplementedError:
            errs.append(
                'One or more unrecognized critical extensions are present in '
                'the CRL entry for the certificate',
                certificate_list_cont,
            )
            raise

    # Step k
    if revoked_reason and revoked_reason.native == 'remove_from_crl':
        revoked_reason = None
        revoked_date = None

    return revoked_date, revoked_reason

async def _handle_single_crl(
    cert: Union[x509.Certificate, cms.AttributeCertificateV2],
    cert_issuer_auth: Authority,
    certificate_list_cont: CRLContainer,
    path: ValidationPath,
    validation_context: ValidationContext,
    delta_lists_by_issuer: Dict[str, List[CRLContainer]],
    use_deltas: bool,
    errs: _CRLErrs,
    proc_state: ValProcState,
) -> Optional[Set[str]]:
    certificate_list = certificate_list_cont.crl_data

    try:
        is_indirect, crl_authority_name = _get_crl_authority_name(
            certificate_list_cont,
            cert_issuer_auth.name,
            certificate_registry=validation_context.certificate_registry,
            errs=errs,
        )
    except LookupError:
        # already logged by _get_crl_authority_name
        return None

    # check if we already know the issuer of this CRL
    crl_issuer = validation_context.revinfo_manager.check_crl_issuer(
        certificate_list
    )
    # if not, attempt to determine it
    if not crl_issuer:
        try:
            crl_issuer_path = await _find_crl_issuer(
                crl_authority_name,
                certificate_list,
                cert=cert,
                cert_issuer_auth=cert_issuer_auth,
                cert_path=path,
                validation_context=validation_context,
                is_indirect=is_indirect,
                proc_state=proc_state,
            )
            crl_issuer = crl_issuer_path.last
        except CRLNoMatchesError:
            # this no-match issue will be dealt with at a higher level later
            errs.issuer_failures += 1
            return None
        except (CertificateFetchError, CRLValidationError) as e:
            errs.append(e.args[0], certificate_list)
            return None

    interim_reasons = _get_crl_scope_assuming_authority(
        crl_issuer=crl_issuer,
        cert=cert,
        certificate_list_cont=certificate_list_cont,
        is_indirect=is_indirect,
        errs=errs,
    )

    if interim_reasons is None:
        return None

    if not _check_crl_freshness(
        certificate_list_cont,
        validation_context.revinfo_policy,
        validation_context.timing_params,
        errs,
        is_delta=False,
    ):
        return None

    # Step c
    if use_deltas:
        delta_certificate_list_cont = _maybe_get_delta_crl(
            certificate_list=certificate_list,
            crl_issuer=crl_issuer,
            policy=validation_context.revinfo_policy,
            timing_params=validation_context.timing_params,
            delta_lists_by_issuer=delta_lists_by_issuer,
            errs=errs,
        )
    else:
        delta_certificate_list_cont = None

    try:
        revoked_date, revoked_reason = _check_cert_on_crl_and_delta(
            crl_issuer=crl_issuer,
            cert=cert,
            certificate_list_cont=certificate_list_cont,
            delta_certificate_list_cont=delta_certificate_list_cont,
            errs=errs,
        )
    except NotImplementedError:
        # the subroutine already registered the failure, so just bail
        return None

    timing = validation_context.timing_params
    control_time = (
        timing.validation_time if timing.point_in_time_validation else None
    )
    if revoked_reason and (control_time is None or revoked_date < control_time):
        raise RevokedError.format(
            reason=revoked_reason,
            revocation_dt=revoked_date,
            revinfo_type='CRL',
            proc_state=proc_state,
        )
    return interim_reasons

def _process_crl_completeness(
    checked_reasons: Set[str],
    total_crls: int,
    errs: _CRLErrs,
    proc_state: ValProcState,
):
    # CRLs should not include this value, but at least one of the examples
    # from the NIST test suite does
    checked_reasons -= {'unused'}

    if checked_reasons != VALID_REVOCATION_REASONS:
        if total_crls == errs.issuer_failures:
            return CRLNoMatchesError(
                f"No CRLs were issued by the issuer of "
                f"{proc_state.describe_cert()}, or any indirect CRL "
                "issuer"
            )

        if not errs.failures:
            errs.append(
                'The available CRLs do not cover all revocation reasons', None
            )

        return CRLValidationIndeterminateError(
            f"Unable to determine if {proc_state.describe_cert()} "
            f"is revoked due to insufficient information from known CRLs",
            failures=errs.failures,
            suspect_stale=(
                errs.stale_last_usable_at
                if errs.freshness_failures_only
                else None
            ),
        )
    
async def verify_crl(
    cert: Union[x509.Certificate, cms.AttributeCertificateV2],
    path: ValidationPath,
    validation_context: ValidationContext,
    use_deltas=True,
    proc_state: Optional[ValProcState] = None,
):
    """
    Verifies a certificate against a list of CRLs, checking to make sure the
    certificate has not been revoked. Uses the algorithm from
    https://tools.ietf.org/html/rfc5280#section-6.3 as a basis, but the
    implementation differs to allow CRLs from unrecorded locations.

    :param cert:
        An asn1crypto.x509.Certificate or asn1crypto.cms.AttributeCertificateV2
        object to check for in the CRLs

    :param path:
        A aurora_certvalidator.path.ValidationPath object of the cert's
        validation path, or in the case of an AC, the AA's validation path.

    :param validation_context:
        A aurora_certvalidator.context.ValidationContext object to use for caching
        validation information

    :param use_deltas:
        A boolean indicating if delta CRLs should be used

    :param proc_state:
        Internal state for error reporting and policy application decisions.

    :raises:
        aurora_certvalidator.errors.CRLNoMatchesError - when none of the CRLs match the certificate
        aurora_certvalidator.errors.CRLValidationError - when any error occurs trying to verify the CertificateList
        aurora_certvalidator.errors.RevokedError - when the CRL indicates the certificate has been revoked
    """

    is_pkc = isinstance(cert, x509.Certificate)
    proc_state = proc_state or ValProcState(
        cert_path_stack=ConsList.sing(path),
        ee_name_override="attribute certificate" if not is_pkc else None,
    )

    revinfo_manager = validation_context.revinfo_manager
    errs = _CRLErrs()
    (
        complete_lists_by_issuer,
        delta_lists_by_issuer,
    ) = await _classify_relevant_crls(revinfo_manager, cert, errs)

    try:
        cert_issuer_auth = path.find_issuing_authority(cert)
    except LookupError:
        raise CRLNoMatchesError(
            f"Could not determine issuer certificate for "
            f"{proc_state.describe_cert()} in path."
        )

    # In the main loop, only complete CRLs are processed, so delta CRLs are
    # weeded out of the to-do list
    crls_to_process = []
    for issuer_crls in complete_lists_by_issuer.values():
        crls_to_process.extend(issuer_crls)
    total_crls = len(crls_to_process)

    checked_reasons = set()

    for certificate_list_cont in crls_to_process:
        try:
            interim_reasons = await _handle_single_crl(
                cert=cert,
                cert_issuer_auth=cert_issuer_auth,
                certificate_list_cont=certificate_list_cont,
                path=path,
                validation_context=validation_context,
                delta_lists_by_issuer=delta_lists_by_issuer,
                use_deltas=use_deltas,
                errs=errs,
                proc_state=proc_state,
            )
            if interim_reasons is not None:
                # Step l
                checked_reasons |= interim_reasons
        except ValueError as e:
            msg = "Generic processing error while validating CRL."
            logging.debug(msg, exc_info=e)
            errs.append(msg, certificate_list_cont)

    exc = _process_crl_completeness(
        checked_reasons, total_crls, errs, proc_state
    )
    if exc is not None:
        raise exc



"""
Asynchronous API for fetching OCSP responses, CRLs and certificates.
"""

import abc
from dataclasses import dataclass
from typing import AsyncGenerator, Iterable, Union

from asn1crypto import cms, crl, ocsp, x509

from .aurora_certvalidator.authority import Authority
from .aurora_certvalidator.version import __version__

__all__ = [
    'OCSPFetcher',
    'CRLFetcher',
    'CertificateFetcher',
    'Fetchers',
    'FetcherBackend'
]

DEFAULT_USER_AGENT = 'aurora_certvalidator %s' % __version__

class OCSPFetcher(abc.ABC):
    """Utility interface to fetch and cache OCSP responses."""

    async def fetch(
        self,
        cert: Union[x509.Certificate, cms.AttributeCertificateV2],
        authority: Authority,
    ) -> ocsp.OCSPResponse:
        """
        Fetch an OCSP response for a certificate.

        :param cert:
            The certificate for which an OCSP response has to be fetched.
        :param authority:
            The issuing authority.
        :raises:
            OCSPFetchError - Raised if an OCSP response could not be obtained.
        :return:
            An OCSP response.
        """
        raise NotImplementedError

    def fetched_responses(self) -> Iterable[ocsp.OCSPResponse]:
        """
        Return all responses fetched by this OCSP fetcher.
        """
        raise NotImplementedError

    def fetched_responses_for_cert(
        self, cert: Union[x509.Certificate, cms.AttributeCertificateV2]
    ) -> Iterable[ocsp.OCSPResponse]:
        """
        Return all responses fetched by this OCSP fetcher that are relevant
        to determine the revocation status of the given certificate.
        """
        raise NotImplementedError


class CRLFetcher(abc.ABC):
    """Utility interface to fetch and cache CRLs."""

    async def fetch(
        self,
        cert: Union[x509.Certificate, cms.AttributeCertificateV2],
        *,
        use_deltas=None,
    ) -> Iterable[crl.CertificateList]:
        """
        Fetches the CRLs for a certificate.

        :param cert:
            An asn1crypto.x509.Certificate object to get the CRL for

        :param use_deltas:
            A boolean indicating if delta CRLs should be fetched

        :raises:
            CRLFetchError - when a network/IO error or decoding error occurs

        :return:
            An iterable of CRLs fetched.
        """
        # side note: we don't want this to be a generator, because in principle,
        #  we always need to consider CRLs from all distribution points together
        #  anyway, so there's no "stream processing" to speak of.
        # (this is currently not 100% efficient in the default implementation,
        #  see comments below)
        raise NotImplementedError

    def fetched_crls(self) -> Iterable[crl.CertificateList]:
        """
        Return all CRLs fetched by this CRL fetcher.
        """
        raise NotImplementedError

    def fetched_crls_for_cert(
        self, cert: Union[x509.Certificate, cms.AttributeCertificateV2]
    ) -> Iterable[crl.CertificateList]:
        """
        Return all relevant fetched CRLs for the given certificate

        :param cert:
            A certificate.
        :return:
            An iterable of CRLs
        :raise KeyError:
            if no fetch operations have been performed for this certificate
        """
        raise NotImplementedError

class CertificateFetcher(abc.ABC):
    """Utility interface to fetch and cache certificates."""

    def fetch_cert_issuers(
        self, cert: Union[x509.Certificate, cms.AttributeCertificateV2]
    ) -> AsyncGenerator[x509.Certificate, None]:
        """
        Fetches certificates from the authority information access extension of
        a certificate.

        :param cert:
            A certificate

        :raises:
            CertificateFetchError - when a network I/O or decoding error occurs

        :return:
            An asynchronous generator yielding asn1crypto.x509.Certificate
            objects that were fetched.
        """
        raise NotImplementedError

    def fetch_crl_issuers(
        self, certificate_list
    ) -> AsyncGenerator[x509.Certificate, None]:
        """
        Fetches certificates from the authority information access extension of
        an asn1crypto.crl.CertificateList.

        :param certificate_list:
            An asn1crypto.crl.CertificateList object

        :raises:
            CertificateFetchError - when a network I/O or decoding error occurs

        :return:
            An asynchronous generator yielding asn1crypto.x509.Certificate
            objects that were fetched.
        """
        raise NotImplementedError

    def fetched_certs(self) -> Iterable[x509.Certificate]:
        """
        Return all certificates retrieved by this certificate fetcher.
        """
        raise NotImplementedError

@dataclass(frozen=True)
class Fetchers:
    """
    Models a collection of fetchers to be used by a validation context.

    The intention is that these can share resources (like a connection pool)
    in a unified, controlled manner. See also :class:`.FetcherBackend`.
    """

    ocsp_fetcher: OCSPFetcher
    crl_fetcher: CRLFetcher
    cert_fetcher: CertificateFetcher

class FetcherBackend(abc.ABC):
    """
    Generic, bare-bones interface to help abstract away instantiation logic for
    fetcher implementations.

    Intended to operate as an asynchronous context manager, with
    `async with backend_obj as fetchers: ...` putting the resulting
    :class:`.Fetchers` object in to the variable named `fetchers`.

    .. note::
        The initialisation part of the API is necessarily synchronous,
        for backwards compatibility with the old ``ValidationContext`` API.
        If you need asynchronous resource management, handle it elsewhere,
        or use some form of lazy resource provisioning.

        Alternatively, you can pass :class:`Fetchers` objects to the validation
        context yourself, and forgo use of the :class:`.FetcherBackend`
        API altogether.
    """

    def get_fetchers(self) -> Fetchers:
        """
        Set up fetchers synchronously.

        .. note::
            This is a synchronous method
        """
        raise NotImplementedError

    async def close(self):
        """
        Clean up the resources associated with this fetcher backend,
        asynchronously.
        """
        pass

    async def __aenter__(self) -> Fetchers:
        return self.get_fetchers()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return await self.close()



"""
This module describes and implements the low-level :class:`.PdfCMSEmbedder`
protocol for embedding CMS payloads into PDF signature objects.
"""

from dataclasses import dataclass
from datetime import datetime
from typing import IO, Optional

from .aurora_signature.pdf_utils import generic, misc
from .aurora_signature.pdf_utils.generic import pdf_name
from .aurora_signature.pdf_utils.layout import BoxConstraints
from .aurora_signature.pdf_utils.writer import BasePdfFileWriter
from .aurora_signature.sign.fields import (
    FieldMDPSpec,
    MDPPerm,
    SigFieldSpec,
    annot_width_height,
    apply_sig_field_spec_properties,
    ensure_sig_flags,
    enumerate_sig_fields,
    get_sig_field_annot,
    prepare_sig_field,
)
from .aurora_signature.sign.general import SigningError
from .aurora_signature.stamp import BaseStampStyle, TextStampStyle

from .pdf_byterange import PdfSignedData

__all__ = [
    'PdfCMSEmbedder',
    'SigMDPSetup',
    'SigObjSetup',
    'SigAppearanceSetup',
    'SigIOSetup',
]


def docmdp_reference_dictionary(permission_level: MDPPerm):
    # this is part of the /Reference entry of the signature object.
    return generic.DictionaryObject(
        {
            pdf_name('/Type'): pdf_name('/SigRef'),
            pdf_name('/TransformMethod'): pdf_name('/DocMDP'),
            pdf_name('/TransformParams'): generic.DictionaryObject(
                {
                    pdf_name('/Type'): pdf_name('/TransformParams'),
                    pdf_name('/V'): pdf_name('/1.2'),
                    pdf_name('/P'): generic.NumberObject(
                        permission_level.value
                    ),
                }
            ),
        }
    )


def fieldmdp_reference_dictionary(
    field_mdp_spec: FieldMDPSpec, data_ref: generic.Reference
):
    data_ind_obj = generic.IndirectObject(
        data_ref.idnum, data_ref.generation, data_ref.pdf
    )
    # this is part of the /Reference entry of the signature object.
    return generic.DictionaryObject(
        {
            pdf_name('/Type'): pdf_name('/SigRef'),
            pdf_name('/TransformMethod'): pdf_name('/FieldMDP'),
            pdf_name('/Data'): data_ind_obj,
            pdf_name('/TransformParams'): field_mdp_spec.as_transform_params(),
        }
    )


# Wrapper around prepare_sig_field with some error reporting


def _get_or_create_sigfield(
    field_name,
    pdf_out: BasePdfFileWriter,
    existing_fields_only,
    new_field_spec: Optional[SigFieldSpec] = None,
):
    root = pdf_out.root
    if field_name is None:
        if not existing_fields_only:
            raise SigningError(
                'Not specifying a field name is only allowed '
                'when existing_fields_only=True'
            )

        # most of the logic in prepare_sig_field has to do with preparing
        # for the potential addition of a new field. That is completely
        # irrelevant in this special case, so we might as well short circuit
        # things.
        field_created = False
        empty_fields = enumerate_sig_fields(pdf_out, filled_status=False)
        try:
            found_field_name, _, sig_field_ref = next(empty_fields)
        except StopIteration:
            raise SigningError('There are no empty signature fields.')

        others = ', '.join(fn for fn, _, _ in empty_fields if fn is not None)
        if others:
            raise SigningError(
                'There are several empty signature fields. Please specify '
                'a field name. The options are %s, %s.'
                % (found_field_name, others)
            )
    else:
        # grab or create a sig field
        if new_field_spec is not None:
            sig_field_kwargs = {
                'box': new_field_spec.box,
                'include_on_page': pdf_out.find_page_for_modification(
                    new_field_spec.on_page
                )[0],
                'combine_annotation': new_field_spec.combine_annotation,
                'invis_settings': new_field_spec.invis_sig_settings,
                'visible_settings': new_field_spec.visible_sig_settings,
            }
        else:
            sig_field_kwargs = {}

        field_created, sig_field_ref = prepare_sig_field(
            field_name,
            root,
            update_writer=pdf_out,
            existing_fields_only=existing_fields_only,
            **sig_field_kwargs,
        )
        if field_created and new_field_spec is not None:
            apply_sig_field_spec_properties(
                pdf_out,
                sig_field=sig_field_ref.get_object(),
                sig_field_spec=new_field_spec,
            )

    ensure_sig_flags(writer=pdf_out, lock_sig_flags=True)

    return field_created, sig_field_ref


@dataclass(frozen=True)
class SigMDPSetup:
    md_algorithm: str
    """
    Message digest algorithm to write into the signature reference dictionary,
    if one is written at all.

    .. warning::
        It is the caller's responsibility to make sure that this value agrees
        with the value embedded into the CMS object, and with the algorithm
        used to hash the document.
        The low-level :class:`.PdfCMSEmbedder` API *will* simply take it at
        face value.
    """

    certify: bool = False
    """
    Sign with an author (certification) signature, as opposed to an approval
    signature. A document can contain at most one such signature, and it must
    be the first one.
    """

    field_lock: Optional[FieldMDPSpec] = None
    """
    Field lock information to write to the signature reference dictionary.
    """

    docmdp_perms: Optional[MDPPerm] = None
    """
    DocMDP permissions to write to the signature reference dictionary.
    """

    def apply(self, sig_obj_ref, writer):
        """
        Apply the settings to a signature object.

        .. danger::
            This method is internal API.
        """

        certify = self.certify
        docmdp_perms = self.docmdp_perms

        lock = self.field_lock

        reference_array = generic.ArrayObject()

        if certify:
            assert docmdp_perms is not None
            # To make a certification signature, we need to leave a record
            #  in the document catalog.
            root = writer.root
            try:
                perms = root['/Perms']
            except KeyError:
                root['/Perms'] = perms = generic.DictionaryObject()
            perms[pdf_name('/DocMDP')] = sig_obj_ref
            writer.update_container(perms)
            reference_array.append(docmdp_reference_dictionary(docmdp_perms))

        if lock is not None:
            fieldmdp_ref = fieldmdp_reference_dictionary(
                lock, data_ref=writer.root_ref
            )
            reference_array.append(fieldmdp_ref)

            if docmdp_perms is not None:
                # NOTE: this is NOT spec-compatible, but emulates Acrobat
                # behaviour
                fieldmdp_ref['/TransformParams']['/P'] = generic.NumberObject(
                    docmdp_perms.value
                )

        if reference_array:
            sig_obj_ref.get_object()['/Reference'] = reference_array


@dataclass(frozen=True)
class SigAppearanceSetup:
    """
    Signature appearance configuration.

    Part of the low-level :class:`.PdfCMSEmbedder` API, see
    :class:`SigObjSetup`.
    """

    style: BaseStampStyle
    """
    Stamp style to use to generate the appearance.
    """

    timestamp: datetime
    """
    Timestamp to show in the signature appearance.
    """

    name: Optional[str]
    """
    Signer name to show in the signature appearance.
    """

    text_params: Optional[dict] = None
    """
    Additional text interpolation parameters to pass to the underlying
    stamp style.
    """

    def apply(self, sig_annot, writer):
        """
        Apply the settings to an annotation.

        .. danger::
            This method is internal API.
        """

        w, h = annot_width_height(sig_annot)
        if w and h:
            # the field is probably a visible one, so we change its appearance
            # stream to show some data about the signature
            stamp = self._appearance_stamp(
                writer, BoxConstraints(width=w, height=h)
            )
            sig_annot['/AP'] = stamp.as_appearances().as_pdf_object()
            try:
                # if there was an entry like this, it's meaningless now
                del sig_annot[pdf_name('/AS')]
            except KeyError:
                pass

    def _appearance_stamp(self, writer, box):
        style = self.style

        name = self.name
        timestamp = self.timestamp
        text_params = {}
        if name is not None:
            text_params['signer'] = name
        text_params.update(self.text_params or {})

        if isinstance(style, TextStampStyle):
            text_params['ts'] = timestamp.strftime(style.timestamp_format)

        return style.create_stamp(writer, box, text_params)


@dataclass(frozen=True)
class SigObjSetup:
    """
    Describes the signature dictionary to be embedded as the form field's value.
    """

    sig_placeholder: PdfSignedData
    """
    Bare-bones placeholder object, usually of type :class:`.SignatureObject`
    or :class:`.DocumentTimestamp`.

    In particular, this determines the number of bytes to allocate for the
    CMS object.
    """

    mdp_setup: Optional[SigMDPSetup] = None
    """
    Optional DocMDP settings, see :class:`.SigMDPSetup`.
    """

    appearance_setup: Optional[SigAppearanceSetup] = None
    """
    Optional appearance settings, see :class:`.SigAppearanceSetup`.
    """


@dataclass(frozen=True)
class SigIOSetup:
    """
    I/O settings for writing signed PDF documents.

    Objects of this type are used in the penultimate phase of
    the :class:`.PdfCMSEmbedder` protocol.
    """

    md_algorithm: str
    """
    Message digest algorithm to use to compute the document hash.
    It should be supported by `pyca/cryptography`.

    .. warning::
        This is also the message digest algorithm that should appear in the
        corresponding ``signerInfo`` entry in the CMS object that ends up
        being embedded in the signature field.
    """

    in_place: bool = False
    """
    Sign the input in-place. If ``False``, write output to a :class:`.BytesIO`
    object, or :attr:`output` if the latter is not ``None``.
    """

    chunk_size: int = misc.DEFAULT_CHUNK_SIZE
    """
    Size of the internal buffer (in bytes) used to feed data to the message 
    digest function if the input stream does not support ``memoryview``.
    """

    output: Optional[IO] = None
    """
    Write the output to the specified output stream. If ``None``, write to a 
    new :class:`.BytesIO` object. Default is ``None``.
    """


class PdfCMSEmbedder:
    """
    Low-level class that handles embedding CMS objects into PDF signature
    fields.

    It also takes care of appearance generation and DocMDP configuration,
    but does not otherwise offer any of the conveniences of
    :class:`.PdfSigner`.

    :param new_field_spec:
        :class:`.SigFieldSpec` to use when creating new fields on-the-fly.
    """

    def __init__(self, new_field_spec: Optional[SigFieldSpec] = None):
        self.new_field_spec = new_field_spec

    def write_cms(
        self,
        field_name: Optional[str],
        writer: BasePdfFileWriter,
        existing_fields_only=False,
    ):
        """
        .. versionadded:: 0.3.0

        .. versionchanged:: 0.7.0
            Digest wrapped in
            :class:`~aurora_signature.sign.signers.pdf_byterange.PreparedByteRangeDigest`
            in step 3; ``output`` returned in step 3 instead of step 4.

        This method returns a generator coroutine that controls the process
        of embedding CMS data into a PDF signature field.
        Can be used for both timestamps and regular signatures.

        .. danger::
            This is a very low-level interface that performs virtually no
            error checking, and is intended to be used in situations
            where the construction of the CMS object to be embedded
            is not under the caller's control (e.g. a remote signer
            that produces full-fledged CMS objects).

            In almost every other case, you're better of using
            :class:`.PdfSigner` instead, with a custom :class:`.Signer`
            implementation to handle the cryptographic operations if necessary.

        The coroutine follows the following specific protocol.

        1. First, it retrieves or creates the signature field to embed the
           CMS object in, and yields a reference to said field.
        2. The caller should then send in a :class:`.SigObjSetup` object, which
           is subsequently processed by the coroutine. For convenience, the
           coroutine will then yield a reference to the signature dictionary
           (as embedded in the PDF writer).
        3. Next, the caller should send a :class:`.SigIOSetup` object,
           describing how the resulting document should be hashed and written
           to the output. The coroutine will write the entire document with a
           placeholder region reserved for the signature and compute the
           document's hash and yield it to the caller.
           It will then yield a ``prepared_digest, output`` tuple, where
           ``prepared_digest`` is a :class:`.PreparedByteRangeDigest` object
           containing the document digest and the relevant offsets, and
           ``output`` is the output stream to which the document to be
           signed was written.

           From this point onwards, **no objects may be changed or added** to
           the :class:`.IncrementalPdfFileWriter` currently in use.
        4. Finally, the caller should pass in a CMS object to place inside
           the signature dictionary. The CMS object can be supplied as a raw
           :class:`bytes` object, or an :mod:`asn1crypto`-style object.
           The coroutine's final yield is the value of the signature
           dictionary's ``/Contents`` entry, given as a hexadecimal string.

        .. caution::
            It is the caller's own responsibility to ensure that enough room
            is available in the placeholder signature object to contain
            the final CMS object.

        :param field_name:
            The name of the field to fill in. This should be a field of type
            ``/Sig``.
        :param writer:
            An :class:`.IncrementalPdfFileWriter` containing the
            document to sign.
        :param existing_fields_only:
            If ``True``, never create a new empty signature field to contain
            the signature.
            If ``False``, a new field may be created if no field matching
            ``field_name`` exists.
        :return:
            A generator coroutine implementing the protocol described above.
        """

        new_field_spec = (
            self.new_field_spec if not existing_fields_only else None
        )
        # start by creating or fetching the appropriate signature field
        field_created, sig_field_ref = _get_or_create_sigfield(
            field_name,
            writer,
            existing_fields_only,
            new_field_spec=new_field_spec,
        )

        # yield control to caller to further process the field dictionary
        # if necessary, request setup specs for sig object
        sig_obj_setup = yield sig_field_ref
        assert isinstance(sig_obj_setup, SigObjSetup)

        sig_field = sig_field_ref.get_object()

        # take care of the field's visual appearance (if applicable)
        appearance_setup = sig_obj_setup.appearance_setup
        if appearance_setup is not None:
            sig_annot = get_sig_field_annot(sig_field)
            appearance_setup.apply(sig_annot, writer)

        sig_obj = sig_obj_setup.sig_placeholder
        sig_obj_ref = writer.add_object(sig_obj)

        # fill in a reference to the (empty) signature object
        sig_field[pdf_name('/V')] = sig_obj_ref

        if not field_created:
            # still need to mark it for updating
            writer.mark_update(sig_field_ref)

        mdp_setup = sig_obj_setup.mdp_setup
        if mdp_setup is not None:
            mdp_setup.apply(sig_obj_ref, writer)

        # again, pass control to the caller
        # and request I/O parameters for putting the cryptographic signature
        # into the output.
        # We pass a reference to the embedded signature object as a convenience.

        sig_io = yield sig_obj_ref
        assert isinstance(sig_io, SigIOSetup)

        # pass control to the sig object's write_signature coroutine
        yield from sig_obj.fill(
            writer,
            sig_io.md_algorithm,
            in_place=sig_io.in_place,
            output=sig_io.output,
            chunk_size=sig_io.chunk_size,
        )



"""Layout utilities (to be expanded)"""

import enum
import logging
from dataclasses import dataclass
from fractions import Fraction
from typing import Optional, Union

from .aurora_signature.config.api import ConfigurableMixin
from .aurora_signature.config.errors import ConfigurationError

__all__ = [
    'LayoutError',
    'BoxSpecificationError',
    'BoxConstraints',
    'AxisAlignment',
    'Margins',
    'InnerScaling',
    'SimpleBoxLayoutRule',
    'Positioning',
]

logger = logging.getLogger(__name__)


class LayoutError(ValueError):
    """Indicates an error in a layout computation."""

    def __init__(self, msg: str, *args):
        self.msg = msg
        super().__init__(msg, *args)


class BoxSpecificationError(LayoutError):
    """Raised when a box constraint is over/underspecified."""

    def __init__(self, msg: Optional[str] = None):
        super().__init__(msg=msg or "box constraint is over/underspecified")


class BoxConstraints:
    """Represents a box of potentially variable width and height.
    Among other uses, this can be leveraged to produce a variably sized
    box with a fixed aspect ratio.

    If width/height are not defined yet, they can be set by assigning to the
    :attr:`width` and :attr:`height` attributes.
    """

    _width: Optional[int]
    _height: Optional[int]
    _ar: Optional[Fraction]
    _fully_specified: bool

    def __init__(
        self,
        width: Union[int, float, None] = None,
        height: Union[int, float, None] = None,
        aspect_ratio: Optional[Fraction] = None,
    ):
        int_width = int(width) if width is not None else None
        int_height = int(height) if height is not None else None
        self._width = int_width
        self._height = int_height

        fully_specified = False

        self._ar = None
        if int_width is None and int_height is None and aspect_ratio is None:
            return
        elif int_width is not None and int_height is not None:
            if aspect_ratio is not None:
                raise BoxSpecificationError  # overspecified
            self._ar = Fraction(int_width, int_height)
            fully_specified = True
        elif aspect_ratio is not None:
            self._ar = aspect_ratio
            if int_height is not None:
                self._width = int(round(int_height * aspect_ratio))
            elif int_width is not None:
                self._height = int(round(int_width / aspect_ratio))

        self._fully_specified = fully_specified

    def _recalculate(self):
        if self._width is not None and self._height is not None:
            self._ar = Fraction(self._width, self._height)
            self._fully_specified = True
        elif self._ar is not None:
            if self._height is not None:
                self._width = int(self._height * self._ar)
                self._fully_specified = True
            elif self._width is not None:
                self._height = int(self._width / self._ar)
                self._fully_specified = True

    @property
    def width(self) -> int:
        """
        :return:
            The width of the box.
        :raises BoxSpecificationError:
            if the box's width could not be determined.
        """
        if self._width is not None:
            return self._width
        else:
            raise BoxSpecificationError

    @width.setter
    def width(self, width):
        if self._width is None:
            self._width = width
            self._recalculate()
        else:
            raise BoxSpecificationError

    @property
    def width_defined(self) -> bool:
        """
        :return:
            ``True`` if the box currently has a well-defined width,
            ``False`` otherwise.
        """
        return self._width is not None

    @property
    def height(self) -> int:
        """
        :return:
            The height of the box.
        :raises BoxSpecificationError:
            if the box's height could not be determined.
        """
        if self._height is not None:
            return self._height
        else:
            raise BoxSpecificationError

    @height.setter
    def height(self, height):
        if self._height is None:
            self._height = height
            self._recalculate()
        else:
            raise BoxSpecificationError

    @property
    def height_defined(self) -> bool:
        """
        :return:
            ``True`` if the box currently has a well-defined height,
            ``False`` otherwise.
        """
        return self._height is not None

    @property
    def aspect_ratio(self) -> Fraction:
        """
        :return:
            The aspect ratio of the box.
        :raises BoxSpecificationError:
            if the box's aspect ratio could not be determined.
        """
        if self._ar is not None:
            return self._ar
        else:
            raise BoxSpecificationError

    @property
    def aspect_ratio_defined(self) -> bool:
        """
        :return:
            ``True`` if the box currently has a well-defined aspect ratio,
            ``False`` otherwise.
        """
        return self._ar is not None


class InnerScaling(enum.Enum):
    """Class representing a scaling convention."""

    NO_SCALING = enum.auto()
    """Never scale content."""

    STRETCH_FILL = enum.auto()
    """Scale content to fill the entire container."""

    STRETCH_TO_FIT = enum.auto()
    """
    Scale content while preserving aspect ratio until either the maximal 
    width or maximal height is reached.
    """

    SHRINK_TO_FIT = enum.auto()
    """
    Scale content down to fit in the container, while preserving the original
    aspect ratio.
    """

    @classmethod
    def from_config(cls, config_str: str) -> 'InnerScaling':
        """
        Convert from a configuration string.

        :param config_str:
            A string: 'none', 'stretch-fill', 'stretch-to-fit', 'shrink-to-fit'
        :return:
            An :class:`.InnerScaling` value.
        :raise ConfigurationError: on unexpected string inputs.
        """
        try:
            return {
                'none': InnerScaling.NO_SCALING,
                'stretch-fill': InnerScaling.STRETCH_FILL,
                'stretch-to-fit': InnerScaling.STRETCH_TO_FIT,
                'shrink-to-fit': InnerScaling.SHRINK_TO_FIT,
            }[config_str.lower()]
        except KeyError:
            raise ConfigurationError(
                f"'{config_str}' is not a valid inner scaling setting; valid "
                f"values are 'none', 'stretch-fill', 'stretch-to-fit', "
                f"'shrink-to-fit'."
            )


class AxisAlignment(enum.Enum):
    """Class representing one-dimensional alignment along an axis."""

    ALIGN_MIN = enum.auto()
    """
    Align maximally towards the negative end of the axis.
    """

    ALIGN_MID = enum.auto()
    """
    Center content along the axis.
    """

    ALIGN_MAX = enum.auto()
    """
    Align maximally towards the positive end of the axis.
    """

    @classmethod
    def from_x_align(cls, align_str: str) -> 'AxisAlignment':
        """
        Convert from a horizontal alignment config string.

        :param align_str:
            A string: 'left', 'mid' or 'right'.
        :return:
            An :class:`.AxisAlignment` value.
        :raise ConfigurationError: on unexpected string inputs.
        """
        try:
            return {
                'left': AxisAlignment.ALIGN_MIN,
                'mid': AxisAlignment.ALIGN_MID,
                'right': AxisAlignment.ALIGN_MAX,
            }[align_str.lower()]
        except KeyError:
            raise ConfigurationError(
                f"'{align_str}' is not a valid horizontal alignment; valid "
                f"values are 'left', 'mid', 'right'."
            )

    @classmethod
    def from_y_align(cls, align_str: str) -> 'AxisAlignment':
        """
        Convert from a vertical alignment config string.

        :param align_str:
            A string: 'bottom', 'mid' or 'top'.
        :return:
            An :class:`.AxisAlignment` value.
        :raise ConfigurationError: on unexpected string inputs.
        """
        try:
            return {
                'bottom': AxisAlignment.ALIGN_MIN,
                'mid': AxisAlignment.ALIGN_MID,
                'top': AxisAlignment.ALIGN_MAX,
            }[align_str.lower()]
        except KeyError:
            raise ConfigurationError(
                f"'{align_str}' is not a valid vertical alignment; valid "
                f"values are 'bottom', 'mid', 'top'."
            )

    @property
    def flipped(self):
        return _alignment_opposites[self]

    def align(
        self, container_len: int, inner_len: int, pre_margin, post_margin
    ) -> int:
        effective_max_len = Margins.effective(
            'length', container_len, pre_margin, post_margin
        )

        if self == AxisAlignment.ALIGN_MAX:
            # we want to start as far up the axis as possible.
            # Ignoring margins, that would be at container_len - inner_len
            # This computation makes sure that there's room for post_margin
            # in the back.
            return container_len - inner_len - post_margin
        elif self == AxisAlignment.ALIGN_MIN:
            return pre_margin
        elif inner_len > effective_max_len:
            logger.warning(
                f"Content box width/height {inner_len} is too wide for "
                f"container size {container_len} with margins "
                f"({pre_margin}, {post_margin}); post_margin will be ignored"
            )
            return pre_margin
        elif self == AxisAlignment.ALIGN_MID:
            # we'll center the inner content *within* the margins
            inner_offset = (effective_max_len - inner_len) // 2
            return pre_margin + inner_offset
        raise TypeError


# Class variables in enums are weird, so let's put this here
_alignment_opposites = {
    AxisAlignment.ALIGN_MID: AxisAlignment.ALIGN_MID,
    AxisAlignment.ALIGN_MIN: AxisAlignment.ALIGN_MAX,
    AxisAlignment.ALIGN_MAX: AxisAlignment.ALIGN_MIN,
}


@dataclass(frozen=True)
class Positioning(ConfigurableMixin):
    """
    Class describing the position and scaling of an object in a container.
    """

    x_pos: int
    """Horizontal coordinate"""

    y_pos: int
    """Vertical coordinate"""

    x_scale: float
    """Horizontal scaling"""

    y_scale: float
    """Vertical scaling"""

    def as_cm(self):
        """
        Convenience method to convert this :class:`.Positioning` into a PDF
        ``cm`` operator.

        :return:
            A byte string representing the ``cm`` operator corresponding
            to this :class:`.Positioning`.
        """
        return b'%g 0 0 %g %g %g cm' % (
            self.x_scale,
            self.y_scale,
            self.x_pos,
            self.y_pos,
        )


def _aln_width(
    alignment: AxisAlignment,
    container_box: BoxConstraints,
    inner_nat_width: int,
    pre_margin: int,
    post_margin: int,
):
    if container_box.width_defined:
        return alignment.align(
            container_box.width, inner_nat_width, pre_margin, post_margin
        )
    else:
        container_box.width = inner_nat_width + pre_margin + post_margin
        return pre_margin


def _aln_height(
    alignment: AxisAlignment,
    container_box: BoxConstraints,
    inner_nat_height: int,
    pre_margin: int,
    post_margin: int,
):
    if container_box.height_defined:
        return alignment.align(
            container_box.height, inner_nat_height, pre_margin, post_margin
        )
    else:
        container_box.height = inner_nat_height + pre_margin + post_margin
        return pre_margin


@dataclass(frozen=True)
class Margins(ConfigurableMixin):
    """Class describing a set of margins."""

    left: int = 0

    right: int = 0

    top: int = 0

    bottom: int = 0

    @classmethod
    def uniform(cls, num):
        """
        Return a set of uniform margins.

        :param num:
            The uniform margin to apply to all four sides.
        :return:
            ``Margins(num, num, num, num)``
        """
        return Margins(num, num, num, num)

    @staticmethod
    def effective(dim_name, container_len, pre, post):
        """Internal helper method to compute effective margins."""
        eff = container_len - pre - post
        if eff < 0:
            raise LayoutError(
                f"Margins ({pre}, {post}) too wide for container "
                f"{dim_name} {container_len}."
            )
        return eff

    def effective_width(self, width):
        """
        Compute width without margins.

        :param width:
            The container width.
        :return:
            The width after subtracting the left and right margins.
        :raises LayoutError:
            if the container width is too short to accommodate the margins.
        """
        return Margins.effective('width', width, self.left, self.right)

    def effective_height(self, height):
        """
        Compute height without margins.

        :param height:
            The container height.
        :return:
            The height after subtracting the top and bottom margins.
        :raises LayoutError:
            if the container height is too short to accommodate the margins.
        """
        return Margins.effective('height', height, self.bottom, self.top)

    @classmethod
    def from_config(cls, config_dict):
        # convenience
        if isinstance(config_dict, list):
            config_dict = dict(
                zip(("left", "right", "top", "bottom"), config_dict)
            )
        return super().from_config(config_dict)


@dataclass(frozen=True)
class SimpleBoxLayoutRule(ConfigurableMixin):
    """
    Class describing alignment, scaling and margin rules for a box
    positioned inside another box.
    """

    x_align: AxisAlignment
    """
    Horizontal alignment settings.
    """

    y_align: AxisAlignment
    """
    Vertical alignment settings.
    """

    margins: Margins = Margins()
    """
    Container (inner) margins. Defaults to all zeroes.
    """

    inner_content_scaling: InnerScaling = InnerScaling.SHRINK_TO_FIT
    """
    Inner content scaling rule.
    """

    @classmethod
    def process_entries(cls, config_dict):
        # in config processing, we default to MID for everything
        x_align = config_dict.get('x_align', AxisAlignment.ALIGN_MID)
        if isinstance(x_align, str):
            x_align = AxisAlignment.from_x_align(x_align)
        config_dict['x_align'] = x_align

        y_align = config_dict.get('y_align', AxisAlignment.ALIGN_MID)
        if isinstance(y_align, str):
            y_align = AxisAlignment.from_y_align(y_align)
        config_dict['y_align'] = y_align

        scaling = config_dict.get('inner_content_scaling', None)
        if scaling is not None:
            config_dict['inner_content_scaling'] = InnerScaling.from_config(
                scaling
            )

    def substitute_margins(self, new_margins: Margins) -> 'SimpleBoxLayoutRule':
        return SimpleBoxLayoutRule(
            x_align=self.x_align,
            y_align=self.y_align,
            margins=new_margins,
            inner_content_scaling=self.inner_content_scaling,
        )

    def fit(
        self,
        container_box: BoxConstraints,
        inner_nat_width: int,
        inner_nat_height: int,
    ) -> Positioning:
        """
        Position and possibly scale a box within a container, according
        to this layout rule.

        :param container_box:
            :class:`.BoxConstraints` describing the container.
        :param inner_nat_width:
            The inner box's natural width.
        :param inner_nat_height:
            The inner box's natural height.
        :return:
            A :class:`.Positioning` describing the scaling & position of the
            lower left corner of the inner box.
        """

        margins = self.margins
        scaling = self.inner_content_scaling
        x_scale = y_scale = 1
        if (
            scaling != InnerScaling.NO_SCALING
            and container_box.width_defined
            and container_box.height_defined
        ):
            eff_width = margins.effective_width(container_box.width)
            eff_height = margins.effective_height(container_box.height)

            x_scale = (
                (eff_width / inner_nat_width) if inner_nat_width != 0 else 1
            )
            y_scale = (
                (eff_height / inner_nat_height) if inner_nat_height != 0 else 1
            )
            if scaling == InnerScaling.STRETCH_TO_FIT:
                x_scale = y_scale = min(x_scale, y_scale)
            elif scaling == InnerScaling.SHRINK_TO_FIT:
                # same as stretch to fit, with the additional stipulation
                # that it can't scale up, only down.
                x_scale = y_scale = min(x_scale, y_scale, 1)

        x_pos = _aln_width(
            self.x_align,
            container_box,
            inner_nat_width * x_scale,
            margins.left,
            margins.right,
        )
        y_pos = _aln_height(
            self.y_align,
            container_box,
            inner_nat_height * y_scale,
            margins.bottom,
            margins.top,
        )
        return Positioning(
            x_pos=x_pos, y_pos=y_pos, x_scale=x_scale, y_scale=y_scale
        )
